# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2011-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import xml.sax as sax
import os
import io
import shutil

from firewall import config
from firewall.core.io.io_object import (
    IO_Object,
    IO_Object_ContentHandler,
    IO_Object_XMLGenerator,
    check_port,
    check_tcpudp,
    check_protocol,
    check_address,
)
from firewall.core.logger import log
from firewall import errors
from firewall.errors import FirewallError


class Service(IO_Object):
    IMPORT_EXPORT_STRUCTURE = {
        "version": "",
        "short": "",
        "description": "",
        "ports": [("", "")],
        "modules": [""],
        "destination": {"": ""},
        "protocols": [""],
        "source_ports": [("", "")],
        "includes": [""],
        "helpers": [""],
    }
    ADDITIONAL_ALNUM_CHARS = ["_", "-"]
    PARSER_REQUIRED_ELEMENT_ATTRS = {
        "short": None,
        "description": None,
        "service": None,
    }
    PARSER_OPTIONAL_ELEMENT_ATTRS = {
        "service": ["name", "version"],
        "port": ["port", "protocol"],
        "protocol": ["value"],
        "module": ["name"],
        "destination": ["ipv4", "ipv6"],
        "source-port": ["port", "protocol"],
        "include": ["service"],
        "helper": ["name"],
    }

    def __init__(self):
        super(Service, self).__init__()
        self.version = ""
        self.short = ""
        self.description = ""
        self.ports = []
        self.protocols = []
        self.modules = []
        self.destination = {}
        self.source_ports = []
        self.includes = []
        self.helpers = []

    def cleanup(self):
        self.version = ""
        self.short = ""
        self.description = ""
        del self.ports[:]
        del self.protocols[:]
        del self.modules[:]
        self.destination.clear()
        del self.source_ports[:]
        del self.includes[:]
        del self.helpers[:]

    def _check_config(self, config, item, all_config, all_io_objects):
        if item == "ports":
            for port in config:
                if port[0] != "":
                    check_port(port[0])
                    check_tcpudp(port[1])
                else:
                    # only protocol
                    check_protocol(port[1])

        elif item == "protocols":
            for proto in config:
                check_protocol(proto)

        elif item == "source_ports":
            for port in config:
                check_port(port[0])
                check_tcpudp(port[1])

        elif item == "destination":
            for destination in config:
                if destination not in ["ipv4", "ipv6"]:
                    raise FirewallError(
                        errors.INVALID_DESTINATION,
                        "'%s' not in {'ipv4'|'ipv6'}" % destination,
                    )
                check_address(destination, config[destination])

        elif item == "modules":
            for module in config:
                if module.startswith("nf_conntrack_"):
                    module = module.replace("nf_conntrack_", "")
                    if "_" in module:
                        module = module.replace("_", "-")
                if len(module) < 2:
                    raise FirewallError(errors.INVALID_MODULE, module)

        elif item == "includes":
            for include in config:
                if include not in all_io_objects["services"]:
                    raise FirewallError(
                        errors.INVALID_SERVICE,
                        "Service '{}': Included service '{}' not found.".format(
                            self.name, include
                        ),
                    )


# PARSER


class service_ContentHandler(IO_Object_ContentHandler):
    def startElement(self, name, attrs):
        IO_Object_ContentHandler.startElement(self, name, attrs)
        self.item.parser_check_element_attrs(name, attrs)
        if name == "service":
            if "name" in attrs:
                log.warning("Ignoring deprecated attribute name='%s'", attrs["name"])
            if "version" in attrs:
                self.item.version = attrs["version"]
        elif name == "short":
            pass
        elif name == "description":
            pass
        elif name == "port":
            if attrs["port"] != "":
                check_port(attrs["port"])
                check_tcpudp(attrs["protocol"])
                entry = (attrs["port"], attrs["protocol"])
                if entry not in self.item.ports:
                    self.item.ports.append(entry)
                else:
                    log.warning(
                        "Port '%s/%s' already set, ignoring.",
                        attrs["port"],
                        attrs["protocol"],
                    )
            else:
                check_protocol(attrs["protocol"])
                if attrs["protocol"] not in self.item.protocols:
                    self.item.protocols.append(attrs["protocol"])
                else:
                    log.warning(
                        "Protocol '%s' already set, ignoring.", attrs["protocol"]
                    )
        elif name == "protocol":
            check_protocol(attrs["value"])
            if attrs["value"] not in self.item.protocols:
                self.item.protocols.append(attrs["value"])
            else:
                log.warning("Protocol '%s' already set, ignoring.", attrs["value"])
        elif name == "source-port":
            check_port(attrs["port"])
            check_tcpudp(attrs["protocol"])
            entry = (attrs["port"], attrs["protocol"])
            if entry not in self.item.source_ports:
                self.item.source_ports.append(entry)
            else:
                log.warning(
                    "SourcePort '%s/%s' already set, ignoring.",
                    attrs["port"],
                    attrs["protocol"],
                )
        elif name == "destination":
            for x in ["ipv4", "ipv6"]:
                if x in attrs:
                    check_address(x, attrs[x])
                    if x in self.item.destination:
                        log.warning(
                            "Destination address for '%s' already set, ignoring", x
                        )
                    else:
                        self.item.destination[x] = attrs[x]
        elif name == "module":
            module = attrs["name"]
            if module.startswith("nf_conntrack_"):
                module = module.replace("nf_conntrack_", "")
                if "_" in module:
                    module = module.replace("_", "-")
            if module not in self.item.modules:
                self.item.modules.append(module)
            else:
                log.warning("Module '%s' already set, ignoring.", module)
        elif name == "include":
            if attrs["service"] not in self.item.includes:
                self.item.includes.append(attrs["service"])
            else:
                log.warning("Include '%s' already set, ignoring.", attrs["service"])
        elif name == "helper":
            if attrs["name"] not in self.item.helpers:
                self.item.helpers.append(attrs["name"])
            else:
                log.warning("Helper '%s' already set, ignoring.", attrs["name"])


def service_reader(filename, path):
    service = Service()
    if not filename.endswith(".xml"):
        raise FirewallError(
            errors.INVALID_NAME, "'%s' is missing .xml suffix" % filename
        )
    service.name = filename[:-4]
    service.check_name(service.name)
    service.filename = filename
    service.path = path
    service.builtin = False if path.startswith(config.ETC_FIREWALLD) else True
    service.default = service.builtin
    handler = service_ContentHandler(service)
    parser = sax.make_parser()
    parser.setContentHandler(handler)
    name = "%s/%s" % (path, filename)
    with open(name, "rb") as f:
        source = sax.InputSource(None)
        source.setByteStream(f)
        try:
            parser.parse(source)
        except sax.SAXParseException as msg:
            raise FirewallError(
                errors.INVALID_SERVICE,
                "not a valid service file: %s" % msg.getException(),
            )
    del handler
    del parser
    return service


def service_writer(service, path=None):
    _path = path if path else service.path

    if service.filename:
        name = "%s/%s" % (_path, service.filename)
    else:
        name = "%s/%s.xml" % (_path, service.name)

    if os.path.exists(name):
        try:
            shutil.copy2(name, "%s.old" % name)
        except Exception as msg:
            log.error("Backup of file '%s' failed: %s", name, msg)

    dirpath = os.path.dirname(name)
    if dirpath.startswith(config.ETC_FIREWALLD) and not os.path.exists(dirpath):
        if not os.path.exists(config.ETC_FIREWALLD):
            os.mkdir(config.ETC_FIREWALLD, 0o750)
        os.mkdir(dirpath, 0o750)

    f = io.open(name, mode="wt", encoding="UTF-8")
    handler = IO_Object_XMLGenerator(f)
    handler.startDocument()

    # start service element
    attrs = {}
    if service.version and service.version != "":
        attrs["version"] = service.version
    handler.startElement("service", attrs)
    handler.ignorableWhitespace("\n")

    # short
    if service.short and service.short != "":
        handler.ignorableWhitespace("  ")
        handler.startElement("short", {})
        handler.characters(service.short)
        handler.endElement("short")
        handler.ignorableWhitespace("\n")

    # description
    if service.description and service.description != "":
        handler.ignorableWhitespace("  ")
        handler.startElement("description", {})
        handler.characters(service.description)
        handler.endElement("description")
        handler.ignorableWhitespace("\n")

    # ports
    for port in service.ports:
        handler.ignorableWhitespace("  ")
        handler.simpleElement("port", {"port": port[0], "protocol": port[1]})
        handler.ignorableWhitespace("\n")

    # protocols
    for protocol in service.protocols:
        handler.ignorableWhitespace("  ")
        handler.simpleElement("protocol", {"value": protocol})
        handler.ignorableWhitespace("\n")

    # source ports
    for port in service.source_ports:
        handler.ignorableWhitespace("  ")
        handler.simpleElement("source-port", {"port": port[0], "protocol": port[1]})
        handler.ignorableWhitespace("\n")

    # modules
    for module in service.modules:
        handler.ignorableWhitespace("  ")
        handler.simpleElement("module", {"name": module})
        handler.ignorableWhitespace("\n")

    # destination
    if len(service.destination) > 0:
        handler.ignorableWhitespace("  ")
        handler.simpleElement("destination", service.destination)
        handler.ignorableWhitespace("\n")

    # includes
    for include in service.includes:
        handler.ignorableWhitespace("  ")
        handler.simpleElement("include", {"service": include})
        handler.ignorableWhitespace("\n")

    # helpers
    for helper in service.helpers:
        handler.ignorableWhitespace("  ")
        handler.simpleElement("helper", {"name": helper})
        handler.ignorableWhitespace("\n")

    # end service element
    handler.endElement("service")
    handler.ignorableWhitespace("\n")
    handler.endDocument()
    f.close()
    del handler



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

"""helper backend"""

from firewall import errors
from firewall.errors import FirewallError


class FirewallHelper:
    def __init__(self, fw):
        self._fw = fw
        self._helpers = {}

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self._helpers)

    # helpers

    def cleanup(self):
        self._helpers.clear()

    def check_helper(self, name):
        if name not in self.get_helpers():
            raise FirewallError(errors.INVALID_HELPER, name)

    def query_helper(self, name):
        return name in self.get_helpers()

    def get_helpers(self):
        return sorted(self._helpers.keys())

    def has_helpers(self):
        return len(self._helpers) > 0

    def get_helper(self, name):
        self.check_helper(name)
        return self._helpers[name]

    def add_helper(self, obj):
        self._helpers[obj.name] = obj

    def remove_helper(self, name):
        if name not in self._helpers:
            raise FirewallError(errors.INVALID_HELPER, name)
        del self._helpers[name]



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2010-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>
#
# signal handling and run_server derived from setroubleshoot
# Copyright (C) 2006,2007,2008,2009 Red Hat, Inc.
# Authors:
#   John Dennis <jdennis@redhat.com>
#   Thomas Liu  <tliu@redhat.com>
#   Dan Walsh <dwalsh@redhat.com>

import signal

from gi.repository import GLib

from firewall.core.logger import log
from firewall.server.firewalld import FirewallD

############################################################################
#
# signal handlers
#
############################################################################


def sighup(service):
    service.reload()
    return True


def sigterm(mainloop):
    mainloop.quit()


############################################################################
#
# run_server function
#
############################################################################


def run_server(debug_gc=False):
    """Main function for firewall server. Handles D-Bus and GLib mainloop."""
    service = None
    if debug_gc:
        from pprint import pformat
        import gc

        gc.enable()
        gc.set_debug(gc.DEBUG_LEAK)

        gc_timeout = 10

        def gc_collect():
            gc.collect()
            if len(gc.garbage) > 0:
                print(
                    "\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"
                    ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n"
                )
                print("GARBAGE OBJECTS (%d):\n" % len(gc.garbage))
                for x in gc.garbage:
                    print(
                        type(x),
                        "\n  ",
                    )
                    print(pformat(x))
                print(
                    "\n<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"
                    "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n"
                )
            GLib.timeout_add_seconds(gc_timeout, gc_collect)

    try:
        service = FirewallD()

        mainloop = GLib.MainLoop()
        if debug_gc:
            GLib.timeout_add_seconds(gc_timeout, gc_collect)

        # use unix_signal_add if available, else unix_signal_add_full
        if hasattr(GLib, "unix_signal_add"):
            unix_signal_add = GLib.unix_signal_add
        else:
            unix_signal_add = GLib.unix_signal_add_full

        unix_signal_add(GLib.PRIORITY_HIGH, signal.SIGHUP, sighup, service)
        unix_signal_add(GLib.PRIORITY_HIGH, signal.SIGTERM, sigterm, mainloop)

        mainloop.run()

    except KeyboardInterrupt:
        log.debug1("Stopping..")

    except SystemExit:
        log.error("Raising SystemExit in run_server")

    except Exception as e:
        log.error("Exception %s: %s", e.__class__.__name__, str(e))

    if service:
        service.stop()



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2012-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

"""This module contains decorators for use with and without D-Bus"""

import dbus
import dbus.service
import traceback
import functools
import inspect
from dbus.exceptions import DBusException

from firewall.errors import FirewallError
from firewall import errors
from firewall.core.logger import log
from firewall.server.dbus import FirewallDBusException, NotAuthorizedException
from firewall.dbus_utils import uid_of_sender

############################################################################
#
# Exception handler decorators
#
############################################################################


def handle_exceptions(func):
    """Decorator to handle exceptions and log them. Used if not conneced
    to D-Bus.
    """

    @functools.wraps(func)
    def _impl(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except FirewallError as error:
            log.debug1(traceback.format_exc())
            log.error(error)
        except Exception:  # pylint: disable=W0703
            log.exception()

    return _impl


def dbus_handle_exceptions(func):
    """Decorator to handle exceptions, log and report them into D-Bus

    :Raises DBusException: on a firewall error code problems.
    """

    @functools.wraps(func)
    def _impl(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except FirewallError as error:
            code = FirewallError.get_code(str(error))
            if code in [
                errors.ALREADY_ENABLED,
                errors.NOT_ENABLED,
                errors.ZONE_ALREADY_SET,
                errors.ALREADY_SET,
            ]:
                log.warning(str(error))
            else:
                log.debug1(traceback.format_exc())
                log.error(str(error))
            raise FirewallDBusException(str(error))
        except DBusException as ex:
            # only log DBusExceptions once
            raise ex
        except Exception as ex:
            log.exception()
            raise FirewallDBusException(str(ex))

    # HACK: functools.wraps() does not copy the function signature and
    # dbus-python doesn't support varargs. As such we need to copy the
    # signature from the function to the newly decorated function otherwise the
    # decorators in dbus-python will manipulate the arg stack and fail
    # miserably.
    #
    # Note: This can be removed if we ever stop using dbus-python.
    #
    # Ref: https://gitlab.freedesktop.org/dbus/dbus-python/-/issues/12
    #
    _impl.__signature__ = inspect.signature(func)
    return _impl


def dbus_service_method(
    dbus_interface,
    in_signature="",
    out_signature="",
    *args,
    is_deprecated=False,
    sender_keyword="sender",
    **kwargs,
):
    def decorator(func):
        if is_deprecated:
            dbus_service_method_deprecated.register(dbus_interface, func.__name__)

        dbus_decorator = dbus.service.method(
            dbus_interface,
            in_signature=in_signature,
            out_signature=out_signature,
            *args,
            sender_keyword=sender_keyword,
            **kwargs,
        )
        return dbus_decorator(func)

    return decorator


def dbus_service_signal(
    dbus_interface,
    signature="",
    *args,
    is_deprecated=False,
    **kwargs,
):
    def decorator(func):
        if is_deprecated:
            dbus_service_signal_deprecated.register(dbus_interface, func.__name__)

        dbus_decorator = dbus.service.signal(
            dbus_interface,
            *args,
            signature=signature,
            **kwargs,
        )
        return dbus_decorator(func)

    return decorator


class dbus_service_deprecated:
    """Decorator that maintains a list of deprecated methods in dbus
    interfaces.
    """

    def __init__(self, interface):
        self.interface = interface

    def __call__(self, func):
        self.register(self.interface, func.__name__)
        return func

    @classmethod
    def register(cls, interface, name):
        s = cls.deprecated.get(interface)
        if s is None:
            s = set()
            cls.deprecated[interface] = s
        s.add(name)


class dbus_service_method_deprecated(dbus_service_deprecated):
    """Decorator that maintains a list of deprecated methods in dbus
    interfaces.
    """

    deprecated = {}


class dbus_service_signal_deprecated(dbus_service_deprecated):
    """Decorator that maintains a list of deprecated signals in dbus
    interfaces.
    """

    deprecated = {}


class dbus_polkit_require_auth:
    """Decorator factory that checks if the interface/method can be used by the
    sender/user. Assumes wrapped function is a method inside a class derived
    from DbusServiceObject.
    """

    _polkit_name = "org.freedesktop.PolicyKit1"
    _polkit_path = "/org/freedesktop/PolicyKit1/Authority"
    _polkit_interface = "org.freedesktop.PolicyKit1.Authority"

    _bus = None
    _bus_signal_receiver = None
    _interface_polkit = None

    def __init__(self, polkit_auth_required):
        self._polkit_auth_required = polkit_auth_required

    @classmethod
    def _polkit_name_owner_changed(cls, name, old_owner, new_owner):
        cls._bus.remove_signal_receiver(cls._bus_signal_receiver)
        cls._bus_signal_receiver = None
        cls._interface_polkit = None

    def __call__(self, func):
        @functools.wraps(func)
        def _impl(*args, **kwargs):
            if not type(self)._bus:
                type(self)._bus = dbus.SystemBus()

            if not type(self)._bus_signal_receiver:
                type(self)._bus_signal_receiver = type(self)._bus.add_signal_receiver(
                    handler_function=type(self)._polkit_name_owner_changed,
                    signal_name="NameOwnerChanged",
                    dbus_interface="org.freedesktop.DBus",
                    arg0=self._polkit_name,
                )

            if not type(self)._interface_polkit:
                try:
                    type(self)._interface_polkit = dbus.Interface(
                        type(self)._bus.get_object(
                            type(self)._polkit_name, type(self)._polkit_path
                        ),
                        type(self)._polkit_interface,
                    )
                except dbus.DBusException:
                    # polkit must not be available
                    pass

            action_id = self._polkit_auth_required
            if not action_id:
                raise dbus.DBusException("Not Authorized: No action_id specified.")

            sender = kwargs.get("sender")
            if sender:
                # use polkit if it's available
                if type(self)._interface_polkit:
                    (result, _, _) = type(self)._interface_polkit.CheckAuthorization(
                        ("system-bus-name", {"name": sender}),
                        action_id,
                        {},
                        1,
                        "",
                        timeout=60,
                    )
                    if not result:
                        raise NotAuthorizedException(action_id, "polkit")
                # fallback to checking UID
                else:
                    uid = uid_of_sender(type(self)._bus, sender)

                    if uid != 0:
                        raise NotAuthorizedException(action_id, "uid")

            return func(*args, **kwargs)

        _impl._polkit_auth_required = self._polkit_auth_required
        return _impl



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2012 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

# fix xmlplus to be compatible with the python xml sax parser and python 3
# by adding __contains__ to xml.sax.xmlreader.AttributesImpl
import xml

if "_xmlplus" in xml.__file__:
    from xml.sax.xmlreader import AttributesImpl

    if not hasattr(AttributesImpl, "__contains__"):
        # this is missing:
        def __AttributesImpl__contains__(self, name):
            return name in getattr(self, "_attrs")

        # add it using the name __contains__
        setattr(AttributesImpl, "__contains__", __AttributesImpl__contains__)
    from xml.sax.saxutils import XMLGenerator

    if not hasattr(XMLGenerator, "_write"):
        # this is missing:
        def __XMLGenerator_write(self, text):
            getattr(self, "_out").write(text)

        # add it using the name _write
        setattr(XMLGenerator, "_write", __XMLGenerator_write)



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

"""ipset io XML handler, reader, writer"""

import xml.sax as sax
import os
import io
import shutil

from firewall import config
from firewall.functions import (
    checkIP,
    checkIP6,
    checkIPnMask,
    checkIP6nMask,
    check_mac,
    check_port,
    checkInterface,
    checkProtocol,
)
from firewall.core.io.io_object import (
    IO_Object,
    IO_Object_ContentHandler,
    IO_Object_XMLGenerator,
)
from firewall.core.ipset import IPSET_TYPES, IPSET_CREATE_OPTIONS
from firewall.core.icmp import (
    check_icmp_name,
    check_icmp_type_code,
    check_icmpv6_name,
    check_icmpv6_type_code,
)
from firewall.core.logger import log
from firewall import errors
from firewall.errors import FirewallError


class IPSet(IO_Object):
    IMPORT_EXPORT_STRUCTURE = {
        "version": "",  # s
        "short": "",  # s
        "description": "",  # s
        "type": "",  # s
        "options": {"": ""},  # a{ss}
        "entries": [""],  # as
    }
    DBUS_SIGNATURE = "(ssssa{ss}as)"
    ADDITIONAL_ALNUM_CHARS = ["_", "-", ":", "."]
    PARSER_REQUIRED_ELEMENT_ATTRS = {
        "short": None,
        "description": None,
        "ipset": ["type"],
        "option": ["name"],
        "entry": None,
    }
    PARSER_OPTIONAL_ELEMENT_ATTRS = {
        "ipset": ["version"],
        "option": ["value"],
    }

    def __init__(self):
        super(IPSet, self).__init__()
        self.version = ""
        self.short = ""
        self.description = ""
        self.type = ""
        self.entries = []
        self.options = {}
        self.applied = False

    def cleanup(self):
        self.version = ""
        self.short = ""
        self.description = ""
        self.type = ""
        del self.entries[:]
        self.options.clear()
        self.applied = False

    @staticmethod
    def check_entry(entry, options, ipset_type):
        family = "ipv4"
        if "family" in options:
            if options["family"] == "inet6":
                family = "ipv6"

        if not ipset_type.startswith("hash:"):
            raise FirewallError(
                errors.INVALID_IPSET, "ipset type '%s' not usable" % ipset_type
            )
        flags = ipset_type[5:].split(",")
        items = entry.split(",")

        if len(flags) != len(items) or len(flags) < 1:
            raise FirewallError(
                errors.INVALID_ENTRY,
                "entry '%s' does not match ipset type '%s'" % (entry, ipset_type),
            )

        for i, flag in enumerate(flags):
            item = items[i]

            if flag == "ip":
                if "-" in item and family == "ipv4":
                    # IP ranges only with plain IPs, no masks
                    if i > 1:
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid address '%s' in '%s'[%d]" % (item, entry, i),
                        )
                    splits = item.split("-")
                    if len(splits) != 2:
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid address range '%s' in '%s' for %s (%s)"
                            % (item, entry, ipset_type, family),
                        )
                    for _split in splits:
                        if (family == "ipv4" and not checkIP(_split)) or (
                            family == "ipv6" and not checkIP6(_split)
                        ):
                            raise FirewallError(
                                errors.INVALID_ENTRY,
                                "invalid address '%s' in '%s' for %s (%s)"
                                % (_split, entry, ipset_type, family),
                            )
                else:
                    # IPs with mask only allowed in the first
                    # position of the type
                    if family == "ipv4":
                        if item == "0.0.0.0":
                            raise FirewallError(
                                errors.INVALID_ENTRY,
                                "invalid address '%s' in '%s' for %s (%s)"
                                % (item, entry, ipset_type, family),
                            )
                        if i == 0:
                            ip_check = checkIPnMask
                        else:
                            ip_check = checkIP
                    else:
                        ip_check = checkIP6
                    if not ip_check(item):
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid address '%s' in '%s' for %s (%s)"
                            % (item, entry, ipset_type, family),
                        )
            elif flag == "net":
                if "-" in item:
                    # IP ranges only with plain IPs, no masks
                    splits = item.split("-")
                    if len(splits) != 2:
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid address range '%s' in '%s' for %s (%s)"
                            % (item, entry, ipset_type, family),
                        )
                    # First part can only be a plain IP
                    if (family == "ipv4" and not checkIP(splits[0])) or (
                        family == "ipv6" and not checkIP6(splits[0])
                    ):
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid address '%s' in '%s' for %s (%s)"
                            % (splits[0], entry, ipset_type, family),
                        )
                    # Second part can also have a mask
                    if (family == "ipv4" and not checkIPnMask(splits[1])) or (
                        family == "ipv6" and not checkIP6nMask(splits[1])
                    ):
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid address '%s' in '%s' for %s (%s)"
                            % (splits[1], entry, ipset_type, family),
                        )
                else:
                    # IPs with mask allowed in all positions, but no /0
                    if item.endswith("/0"):
                        if not (
                            family == "ipv6"
                            and i == 0
                            and ipset_type == "hash:net,iface"
                        ):
                            raise FirewallError(
                                errors.INVALID_ENTRY,
                                "invalid address '%s' in '%s' for %s (%s)"
                                % (item, entry, ipset_type, family),
                            )
                    if (family == "ipv4" and not checkIPnMask(item)) or (
                        family == "ipv6" and not checkIP6nMask(item)
                    ):
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid address '%s' in '%s' for %s (%s)"
                            % (item, entry, ipset_type, family),
                        )
            elif flag == "mac":
                # ipset does not allow to add 00:00:00:00:00:00
                if not check_mac(item) or item == "00:00:00:00:00:00":
                    raise FirewallError(
                        errors.INVALID_ENTRY,
                        "invalid mac address '%s' in '%s'" % (item, entry),
                    )
            elif flag == "port":
                if ":" in item:
                    splits = item.split(":")
                    if len(splits) != 2:
                        raise FirewallError(
                            errors.INVALID_ENTRY, "invalid port '%s'" % (item)
                        )
                    if splits[0] == "icmp":
                        if family != "ipv4":
                            raise FirewallError(
                                errors.INVALID_ENTRY,
                                "invalid protocol for family '%s' in '%s'"
                                % (family, entry),
                            )
                        if not check_icmp_name(splits[1]) and "/" not in splits[1]:
                            raise FirewallError(
                                errors.INVALID_ENTRY,
                                "invalid icmp type '%s' in '%s'" % (splits[1], entry),
                            )
                        else:
                            (_type, _code) = splits[1].split("/")
                            if not check_icmp_type_code(_type, _code):
                                raise FirewallError(
                                    errors.INVALID_ENTRY,
                                    "invalid icmp type '%s' in '%s'"
                                    % (splits[1], entry),
                                )
                    elif splits[0] in ["icmpv6", "ipv6-icmp"]:
                        if family != "ipv6":
                            raise FirewallError(
                                errors.INVALID_ENTRY,
                                "invalid protocol for family '%s' in '%s'"
                                % (family, entry),
                            )
                        if not check_icmpv6_name(splits[1]) and "/" not in splits[1]:
                            raise FirewallError(
                                errors.INVALID_ENTRY,
                                "invalid icmpv6 type '%s' in '%s'" % (splits[1], entry),
                            )
                        else:
                            (_type, _code) = splits[1].split("/")
                            if not check_icmpv6_type_code(_type, _code):
                                raise FirewallError(
                                    errors.INVALID_ENTRY,
                                    "invalid icmpv6 type '%s' in '%s'"
                                    % (splits[1], entry),
                                )
                    elif splits[0] not in [
                        "tcp",
                        "sctp",
                        "udp",
                        "udplite",
                    ] and not checkProtocol(splits[0]):
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid protocol '%s' in '%s'" % (splits[0], entry),
                        )
                    elif not check_port(splits[1]):
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid port '%s'in '%s'" % (splits[1], entry),
                        )
                else:
                    if not check_port(item):
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid port '%s' in '%s'" % (item, entry),
                        )
            elif flag == "mark":
                if item.startswith("0x"):
                    try:
                        int_val = int(item, 16)
                    except ValueError:
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid mark '%s' in '%s'" % (item, entry),
                        )
                else:
                    try:
                        int_val = int(item)
                    except ValueError:
                        raise FirewallError(
                            errors.INVALID_ENTRY,
                            "invalid mark '%s' in '%s'" % (item, entry),
                        )
                if int_val < 0 or int_val > 4294967295:
                    raise FirewallError(
                        errors.INVALID_ENTRY,
                        "invalid mark '%s' in '%s'" % (item, entry),
                    )
            elif flag == "iface":
                if not checkInterface(item) or len(item) > 15:
                    raise FirewallError(
                        errors.INVALID_ENTRY,
                        "invalid interface '%s' in '%s'" % (item, entry),
                    )
            else:
                raise FirewallError(
                    errors.INVALID_IPSET, "ipset type '%s' not usable" % ipset_type
                )

    def _check_config(self, config, item, all_config, all_io_objects):
        if item == "type":
            if config not in IPSET_TYPES:
                raise FirewallError(
                    errors.INVALID_TYPE, "'%s' is not valid ipset type" % config
                )
        if item == "options":
            for key in config.keys():
                if key not in IPSET_CREATE_OPTIONS:
                    raise FirewallError(
                        errors.INVALID_IPSET, "ipset invalid option '%s'" % key
                    )
                if key in ["timeout", "hashsize", "maxelem"]:
                    try:
                        int_value = int(config[key])
                    except ValueError:
                        raise FirewallError(
                            errors.INVALID_VALUE,
                            "Option '%s': Value '%s' is not an integer"
                            % (key, config[key]),
                        )
                    if int_value < 0:
                        raise FirewallError(
                            errors.INVALID_VALUE,
                            "Option '%s': Value '%s' is negative" % (key, config[key]),
                        )
                    if key == "maxelem" and len(self.entries) > int_value:
                        raise FirewallError(
                            errors.INVALID_IPSET,
                            "the number of ipset entries exceeds the maximum set size of %d"
                            % int_value,
                        )
                elif key == "family" and config[key] not in ["inet", "inet6"]:
                    raise FirewallError(errors.INVALID_FAMILY, config[key])

    def import_config(self, config, all_io_objects):
        if "timeout" in config[4] and config[4]["timeout"] != "0":
            if len(config[5]) != 0:
                raise FirewallError(errors.IPSET_WITH_TIMEOUT)
        for entry in config[5]:
            IPSet.check_entry(entry, config[4], config[3])
        super(IPSet, self).import_config(config, all_io_objects)


# PARSER


class ipset_ContentHandler(IO_Object_ContentHandler):
    def startElement(self, name, attrs):
        IO_Object_ContentHandler.startElement(self, name, attrs)
        self.item.parser_check_element_attrs(name, attrs)
        if name == "ipset":
            if "type" in attrs:
                if attrs["type"] not in IPSET_TYPES:
                    raise FirewallError(errors.INVALID_TYPE, "%s" % attrs["type"])
                self.item.type = attrs["type"]
            if "version" in attrs:
                self.item.version = attrs["version"]
        elif name == "short":
            pass
        elif name == "description":
            pass
        elif name == "option":
            value = ""
            if "value" in attrs:
                value = attrs["value"]

            if attrs["name"] not in ["family", "timeout", "hashsize", "maxelem"]:
                raise FirewallError(
                    errors.INVALID_OPTION, "Unknown option '%s'" % attrs["name"]
                )
            if self.item.type == "hash:mac" and attrs["name"] in ["family"]:
                raise FirewallError(
                    errors.INVALID_OPTION,
                    "Unsupported option '%s' for type '%s'"
                    % (attrs["name"], self.item.type),
                )
            if (
                attrs["name"] in ["family", "timeout", "hashsize", "maxelem"]
                and not value
            ):
                raise FirewallError(
                    errors.INVALID_OPTION,
                    "Missing mandatory value of option '%s'" % attrs["name"],
                )
            if attrs["name"] in ["timeout", "hashsize", "maxelem"]:
                try:
                    int_value = int(value)
                except ValueError:
                    raise FirewallError(
                        errors.INVALID_VALUE,
                        "Option '%s': Value '%s' is not an integer"
                        % (attrs["name"], value),
                    )
                if int_value < 0:
                    raise FirewallError(
                        errors.INVALID_VALUE,
                        "Option '%s': Value '%s' is negative" % (attrs["name"], value),
                    )
            if attrs["name"] == "family" and value not in ["inet", "inet6"]:
                raise FirewallError(errors.INVALID_FAMILY, value)
            if attrs["name"] not in self.item.options:
                self.item.options[attrs["name"]] = value
            else:
                log.warning("Option %s already set, ignoring.", attrs["name"])
        # nothing to do for entry and entries here

    def endElement(self, name):
        IO_Object_ContentHandler.endElement(self, name)
        if name == "entry":
            self.item.entries.append(self._element)


def ipset_reader(filename, path):
    ipset = IPSet()
    if not filename.endswith(".xml"):
        raise FirewallError(
            errors.INVALID_NAME, "'%s' is missing .xml suffix" % filename
        )
    ipset.name = filename[:-4]
    ipset.check_name(ipset.name)
    ipset.filename = filename
    ipset.path = path
    ipset.builtin = False if path.startswith(config.ETC_FIREWALLD) else True
    ipset.default = ipset.builtin
    handler = ipset_ContentHandler(ipset)
    parser = sax.make_parser()
    parser.setContentHandler(handler)
    name = "%s/%s" % (path, filename)
    with open(name, "rb") as f:
        source = sax.InputSource(None)
        source.setByteStream(f)
        try:
            parser.parse(source)
        except sax.SAXParseException as msg:
            raise FirewallError(
                errors.INVALID_IPSET, "not a valid ipset file: %s" % msg.getException()
            )
    del handler
    del parser
    if (
        "timeout" in ipset.options
        and ipset.options["timeout"] != "0"
        and len(ipset.entries) > 0
    ):
        # no entries visible for ipsets with timeout
        log.warning(
            "ipset '%s': timeout option is set, entries are ignored", ipset.name
        )
        del ipset.entries[:]
    i = 0
    entries_set = set()
    while i < len(ipset.entries):
        if ipset.entries[i] in entries_set:
            log.warning("Entry %s already set, ignoring.", ipset.entries[i])
            ipset.entries.pop(i)
        else:
            try:
                ipset.check_entry(ipset.entries[i], ipset.options, ipset.type)
            except FirewallError as e:
                log.warning("%s, ignoring.", e)
                ipset.entries.pop(i)
            else:
                entries_set.add(ipset.entries[i])
                i += 1
    del entries_set

    return ipset


def ipset_writer(ipset, path=None):
    _path = path if path else ipset.path

    if ipset.filename:
        name = "%s/%s" % (_path, ipset.filename)
    else:
        name = "%s/%s.xml" % (_path, ipset.name)

    if os.path.exists(name):
        try:
            shutil.copy2(name, "%s.old" % name)
        except Exception as msg:
            log.error("Backup of file '%s' failed: %s", name, msg)

    dirpath = os.path.dirname(name)
    if dirpath.startswith(config.ETC_FIREWALLD) and not os.path.exists(dirpath):
        if not os.path.exists(config.ETC_FIREWALLD):
            os.mkdir(config.ETC_FIREWALLD, 0o750)
        os.mkdir(dirpath, 0o750)

    f = io.open(name, mode="wt", encoding="UTF-8")
    handler = IO_Object_XMLGenerator(f)
    handler.startDocument()

    # start ipset element
    attrs = {"type": ipset.type}
    if ipset.version and ipset.version != "":
        attrs["version"] = ipset.version
    handler.startElement("ipset", attrs)
    handler.ignorableWhitespace("\n")

    # short
    if ipset.short and ipset.short != "":
        handler.ignorableWhitespace("  ")
        handler.startElement("short", {})
        handler.characters(ipset.short)
        handler.endElement("short")
        handler.ignorableWhitespace("\n")

    # description
    if ipset.description and ipset.description != "":
        handler.ignorableWhitespace("  ")
        handler.startElement("description", {})
        handler.characters(ipset.description)
        handler.endElement("description")
        handler.ignorableWhitespace("\n")

    # options
    for key, value in ipset.options.items():
        handler.ignorableWhitespace("  ")
        if value != "":
            handler.simpleElement("option", {"name": key, "value": value})
        else:
            handler.simpleElement("option", {"name": key})
        handler.ignorableWhitespace("\n")

    # entries
    for entry in ipset.entries:
        handler.ignorableWhitespace("  ")
        handler.startElement("entry", {})
        handler.characters(entry)
        handler.endElement("entry")
        handler.ignorableWhitespace("\n")

    # end ipset element
    handler.endElement("ipset")
    handler.ignorableWhitespace("\n")
    handler.endDocument()
    f.close()
    del handler



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2017 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

ICMP_TYPES = {
    # "type": (type, code, backend omit code)
    "echo-reply": (0, 0, True),
    "pong": (0, 0, True),
    "destination-unreachable": (3, 0, True),
    "network-unreachable": (3, 0, False),
    "tos-network-unreachable": (3, 0, False),
    "host-unreachable": (3, 1, False),
    "tos-host-unreachable": (3, 1, False),
    "protocol-unreachable": (3, 2, False),
    "port-unreachable": (3, 3, False),
    "fragmentation-needed": (3, 4, False),
    "source-route-failed": (3, 5, False),
    # RFC-1112 Section 3.2.2.1 defines type 3, code 6-12
    "network-unknown": (3, 6, False),
    "host-unknown": (3, 7, False),
    "network-prohibited": (3, 9, False),
    "host-prohibited": (3, 10, False),
    "TOS-network-unreachable": (3, 11, False),
    "TOS-host-unreachable": (3, 12, False),
    # RFC-1812 Section 5.2.7.1 defines type 3, code 13-15
    "communication-prohibited": (3, 13, False),
    "host-precedence-violation": (3, 14, False),
    "precedence-cutoff": (3, 15, False),
    "source-quench": (4, 0, True),
    "network-redirect": (5, 0, False),
    "redirect": (5, 0, True),
    "host-redirect": (5, 1, False),
    "tos-host-redirect": (5, 1, False),
    "TOS-network-redirect": (5, 2, False),
    "tos-network-redirect": (5, 2, False),
    "TOS-host-redirect": (5, 3, False),
    "echo-request": (8, 0, True),
    "ping": (8, 0, True),
    "router-advertisement": (9, 0, True),
    "router-solicitation": (10, 0, True),
    "time-exceeded": (11, 0, True),
    "ttl-zero-during-transit": (11, 0, False),
    "ttl-zero-during-reassembly": (11, 1, False),
    "parameter-problem": (12, 0, True),
    "ip-header-bad": (12, 0, False),
    "required-option-missing": (12, 1, False),
    "timestamp-request": (13, 0, True),
    "timestamp-reply": (14, 0, True),
    "address-mask-request": (17, 0, False),
    "address-mask-reply": (18, 0, False),
}

ICMPV6_TYPES = {
    # "type": (type, code, backend omit code)
    "destination-unreachable": (1, 0, True),
    "no-route": (1, 0, False),
    "communication-prohibited": (1, 1, False),
    "beyond-scope": (1, 2, False),
    "address-unreachable": (1, 3, False),
    "port-unreachable": (1, 4, False),
    "failed-policy": (1, 5, False),
    "reject-route": (1, 6, False),
    "packet-too-big": (2, 0, True),
    "time-exceeded": (3, 0, True),
    "ttl-zero-during-transit": (3, 0, False),
    "ttl-zero-during-reassembly": (3, 1, False),
    "parameter-problem": (4, 0, True),
    "bad-header": (4, 0, False),
    "unknown-header-type": (4, 1, False),
    "unknown-option": (4, 2, False),
    "echo-request": (128, 0, True),
    "ping": (128, 0, True),
    "echo-reply": (129, 0, True),
    "pong": (129, 0, True),
    "router-solicitation": (133, 0, False),
    "router-advertisement": (134, 0, False),
    "neighbour-solicitation": (135, 0, False),
    "neigbour-solicitation": (135, 0, False),
    "neighbour-advertisement": (136, 0, False),
    "neigbour-advertisement": (136, 0, False),
    "redirect": (137, 0, False),
    # MLD is RFC-2710
    "mld-listener-query": (130, 0, True),
    "mld-listener-report": (131, 0, True),
    "mld-listener-done": (132, 0, True),
    # MLDv2 is RFC-9777
    "mld2-listener-report": (143, 0, True),
}


def check_icmp_name(_name):
    if _name in ICMP_TYPES:
        return True
    return False


def check_icmp_type_code(_type, _code):
    if (_type, _code) in ICMP_TYPES.values():
        return True
    return False


def check_icmpv6_name(_name):
    if _name in ICMPV6_TYPES:
        return True
    return False


def check_icmpv6_type_code(_type, _code):
    if (_type, _code) in ICMPV6_TYPES.values():
        return True
    return False



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2011-2012 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import os.path
import io
import tempfile
import shutil

from firewall import config
from firewall.core.logger import log
from firewall import errors

valid_keys = [
    "DefaultZone",
    "MinimalMark",
    "CleanupOnExit",
    "CleanupModulesOnExit",
    "IPv6_rpfilter",
    "IndividualCalls",
    "LogDenied",
    "AutomaticHelpers",
    "FirewallBackend",
    "FlushAllOnReload",
    "ReloadPolicy",
    "RFC3964_IPv4",
    "AllowZoneDrifting",
    "NftablesFlowtable",
    "NftablesCounters",
    "NftablesTableOwner",
    "StrictForwardPorts",
    # Deprecated keys:
    "Lockdown",
]


class firewalld_conf:
    def __init__(self, filename):
        self._config = {}
        self._deleted = []
        self.filename = filename
        self.clear()

    def clear(self):
        self._config = {}
        self._deleted = []

    def cleanup(self):
        self._config.clear()
        self._deleted = []

    def get(self, key):
        return self._config.get(key.strip())

    def set(self, key, value):
        _key = key.strip()
        self._config[_key] = value.strip()
        if _key in self._deleted:
            self._deleted.remove(_key)

    def __str__(self):
        s = ""
        for key, value in self._config.items():
            if s:
                s += "\n"
            s += "%s=%s" % (key, value)
        return s

    def set_defaults(self):
        self.set("DefaultZone", config.FALLBACK_ZONE)
        self.set("MinimalMark", str(config.FALLBACK_MINIMAL_MARK))
        self.set("CleanupOnExit", "yes" if config.FALLBACK_CLEANUP_ON_EXIT else "no")
        self.set(
            "CleanupModulesOnExit",
            "yes" if config.FALLBACK_CLEANUP_MODULES_ON_EXIT else "no",
        )
        self.set("IPv6_rpfilter", config.FALLBACK_IPV6_RPFILTER)
        self.set("IndividualCalls", "yes" if config.FALLBACK_INDIVIDUAL_CALLS else "no")
        self.set("LogDenied", config.FALLBACK_LOG_DENIED)
        self.set("AutomaticHelpers", config.FALLBACK_AUTOMATIC_HELPERS)
        self.set("FirewallBackend", config.FALLBACK_FIREWALL_BACKEND)
        self.set(
            "FlushAllOnReload", "yes" if config.FALLBACK_FLUSH_ALL_ON_RELOAD else "no"
        )
        self.set("ReloadPolicy", config.FALLBACK_RELOAD_POLICY)
        self.set("RFC3964_IPv4", "yes" if config.FALLBACK_RFC3964_IPV4 else "no")
        self.set(
            "AllowZoneDrifting", "yes" if config.FALLBACK_ALLOW_ZONE_DRIFTING else "no"
        )
        self.set("NftablesFlowtable", config.FALLBACK_NFTABLES_FLOWTABLE)
        self.set(
            "NftablesCounters", "yes" if config.FALLBACK_NFTABLES_COUNTERS else "no"
        )
        self.set(
            "NftablesTableOwner",
            "yes" if config.FALLBACK_NFTABLES_TABLE_OWNER else "no",
        )
        self.set(
            "StrictForwardPorts",
            "yes" if config.FALLBACK_STRICT_FORWARD_PORTS else "no",
        )

    def sanity_check(self):
        if self.get("FirewallBackend") == "iptables" and self.get("IPv6_rpfilter") in (
            "loose-forward",
            "strict-forward",
        ):
            raise errors.FirewallError(
                errors.INVALID_VALUE,
                f"IPv6_rpfilter={self.get('IPv6_rpfilter')} is incompatible "
                "with FirewallBackend=iptables. This is a limitation "
                "of the iptables backend.",
            )

    # load self.filename
    def read(self):
        self.clear()
        try:
            f = open(self.filename, "r")
        except Exception as msg:
            log.error("Failed to load '%s': %s", self.filename, msg)
            self.set_defaults()
            raise

        for line in f:
            if not line:
                break
            line = line.strip()
            if len(line) < 1 or line[0] in ["#", ";"]:
                continue
            # get key/value pair
            pair = [x.strip() for x in line.split("=")]
            if len(pair) != 2:
                log.error("Invalid option definition: '%s'", line.strip())
                continue
            elif pair[0] not in valid_keys:
                log.error("Invalid option: '%s'", line.strip())
                continue
            elif pair[1] == "":
                log.error("Missing value: '%s'", line.strip())
                continue
            elif self._config.get(pair[0]) is not None:
                log.error("Duplicate option definition: '%s'", line.strip())
                continue
            self._config[pair[0]] = pair[1]
        f.close()

        # check default zone
        if not self.get("DefaultZone"):
            log.error(
                "DefaultZone is not set, using default value '%s'", config.FALLBACK_ZONE
            )
            self.set("DefaultZone", str(config.FALLBACK_ZONE))

        # check minimal mark
        value = self.get("MinimalMark")
        try:
            int(value)
        except (ValueError, TypeError):
            if value is not None:
                log.warning(
                    "MinimalMark '%s' is not valid, using default " "value '%d'",
                    value if value else "",
                    config.FALLBACK_MINIMAL_MARK,
                )
            self.set("MinimalMark", str(config.FALLBACK_MINIMAL_MARK))

        # check cleanup on exit
        value = self.get("CleanupOnExit")
        if not value or value.lower() not in ["no", "false", "yes", "true"]:
            if value is not None:
                log.warning(
                    "CleanupOnExit '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_CLEANUP_ON_EXIT,
                )
            self.set(
                "CleanupOnExit", "yes" if config.FALLBACK_CLEANUP_ON_EXIT else "no"
            )

        # check module cleanup on exit
        value = self.get("CleanupModulesOnExit")
        if not value or value.lower() not in ["no", "false", "yes", "true"]:
            if value is not None:
                log.warning(
                    "CleanupModulesOnExit '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_CLEANUP_MODULES_ON_EXIT,
                )
            self.set(
                "CleanupModulesOnExit",
                "yes" if config.FALLBACK_CLEANUP_MODULES_ON_EXIT else "no",
            )

        value = self.get("Lockdown")
        if value and value.lower() not in ["no", "false"]:
            log.warning('"Lockdown" option is deprecated and has no effect')

        # check ipv6_rpfilter
        value = self.get("IPv6_rpfilter")
        if not value or value.lower() not in config.IPV6_RPFILTER_VALUES:
            if value is not None:
                log.warning(
                    "IPv6_rpfilter '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_IPV6_RPFILTER,
                )
            self.set("IPv6_rpfilter", config.FALLBACK_IPV6_RPFILTER)

        # check individual calls
        value = self.get("IndividualCalls")
        if not value or value.lower() not in ["yes", "true", "no", "false"]:
            if value is not None:
                log.warning(
                    "IndividualCalls '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_INDIVIDUAL_CALLS,
                )
            self.set(
                "IndividualCalls", "yes" if config.FALLBACK_INDIVIDUAL_CALLS else "no"
            )

        # check log denied
        value = self.get("LogDenied")
        if not value or value not in config.LOG_DENIED_VALUES:
            if value is not None:
                log.warning(
                    "LogDenied '%s' is invalid, using default value '%s'",
                    value,
                    config.FALLBACK_LOG_DENIED,
                )
            self.set("LogDenied", str(config.FALLBACK_LOG_DENIED))

        # check automatic helpers
        value = self.get("AutomaticHelpers")
        if not value or value.lower() not in config.AUTOMATIC_HELPERS_VALUES:
            if value is not None:
                log.warning(
                    "AutomaticHelpers '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_AUTOMATIC_HELPERS,
                )
            self.set("AutomaticHelpers", str(config.FALLBACK_AUTOMATIC_HELPERS))

        value = self.get("FirewallBackend")
        if not value or value.lower() not in config.FIREWALL_BACKEND_VALUES:
            if value is not None:
                log.warning(
                    "FirewallBackend '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_FIREWALL_BACKEND,
                )
            self.set("FirewallBackend", str(config.FALLBACK_FIREWALL_BACKEND))

        value = self.get("FlushAllOnReload")
        if not value or value.lower() not in ["yes", "true", "no", "false"]:
            if value is not None:
                log.warning(
                    "FlushAllOnReload '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_FLUSH_ALL_ON_RELOAD,
                )
            self.set("FlushAllOnReload", str(config.FALLBACK_FLUSH_ALL_ON_RELOAD))

        value = self.get("ReloadPolicy")
        try:
            value = self._parse_reload_policy(value)
        except ValueError:
            log.warning(
                "ReloadPolicy '%s' is not valid, using default value '%s'",
                value,
                config.FALLBACK_RELOAD_POLICY,
            )
            self.set("ReloadPolicy", config.FALLBACK_RELOAD_POLICY)

        value = self.get("RFC3964_IPv4")
        if not value or value.lower() not in ["yes", "true", "no", "false"]:
            if value is not None:
                log.warning(
                    "RFC3964_IPv4 '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_RFC3964_IPV4,
                )
            self.set("RFC3964_IPv4", str(config.FALLBACK_RFC3964_IPV4))

        value = self.get("AllowZoneDrifting")
        if not value or value.lower() not in ["yes", "true", "no", "false"]:
            if value is not None:
                log.warning(
                    "AllowZoneDrifting '%s' is not valid, using default " "value %s",
                    value if value else "",
                    config.FALLBACK_ALLOW_ZONE_DRIFTING,
                )
            self.set(
                "AllowZoneDrifting",
                "yes" if config.FALLBACK_ALLOW_ZONE_DRIFTING else "no",
            )

        self.sanity_check()

    # save to self.filename if there are key/value changes
    def write(self):
        if len(self._config) < 1:
            # no changes: nothing to do
            return

        # handled keys
        done = []

        if not os.path.exists(config.ETC_FIREWALLD):
            os.mkdir(config.ETC_FIREWALLD, 0o750)

        try:
            temp_file = tempfile.NamedTemporaryFile(
                mode="wt",
                prefix="%s." % os.path.basename(self.filename),
                dir=os.path.dirname(self.filename),
                delete=False,
            )
        except Exception as msg:
            log.error("Failed to open temporary file: %s" % msg)
            raise

        modified = False
        empty = False
        try:
            f = io.open(self.filename, mode="rt", encoding="UTF-8")
        except Exception as msg:
            if os.path.exists(self.filename):
                log.error("Failed to open '%s': %s" % (self.filename, msg))
                raise
            else:
                f = None
        else:
            for line in f:
                if not line:
                    break
                # remove newline
                line = line.strip("\n")

                if len(line) < 1:
                    if not empty:
                        temp_file.write("\n")
                        empty = True
                elif line[0] == "#":
                    empty = False
                    temp_file.write(line)
                    temp_file.write("\n")
                else:
                    p = line.split("=")
                    if len(p) != 2:
                        empty = False
                        temp_file.write(line + "\n")
                        continue
                    key = p[0].strip()
                    value = p[1].strip()
                    # check for modified key/value pairs
                    if key not in done:
                        if key in self._config and self._config[key] != value:
                            empty = False
                            temp_file.write("%s=%s\n" % (key, self._config[key]))
                            modified = True
                        elif key in self._deleted:
                            modified = True
                        else:
                            empty = False
                            temp_file.write(line + "\n")
                        done.append(key)
                    else:
                        modified = True

        # write remaining key/value pairs
        if len(self._config) > 0:
            for key, value in self._config.items():
                if key in done:
                    continue
                if key in [
                    "MinimalMark",
                    "AutomaticHelpers",
                    "AllowZoneDrifting",
                ]:  # omit deprecated from new config
                    continue
                if not empty:
                    temp_file.write("\n")
                    empty = True
                temp_file.write("%s=%s\n" % (key, value))
                modified = True

        if f:
            f.close()
        temp_file.close()

        if not modified:  # not modified: remove tempfile
            os.remove(temp_file.name)
            return
        # make backup
        if os.path.exists(self.filename):
            try:
                shutil.copy2(self.filename, "%s.old" % self.filename)
            except Exception as msg:
                os.remove(temp_file.name)
                raise IOError("Backup of '%s' failed: %s" % (self.filename, msg))

        # copy tempfile
        try:
            shutil.move(temp_file.name, self.filename)
        except Exception as msg:
            os.remove(temp_file.name)
            raise IOError("Failed to create '%s': %s" % (self.filename, msg))
        else:
            os.chmod(self.filename, 0o600)

    @staticmethod
    def _parse_reload_policy(value):
        valid = True
        result = {
            "INPUT": "DROP",
            "FORWARD": "DROP",
            "OUTPUT": "DROP",
        }
        if value:
            value = value.strip()
            v = value.upper()
            if v in ("ACCEPT", "REJECT", "DROP"):
                for k in result:
                    result[k] = v
            else:
                for a in value.replace(";", ",").split(","):
                    a = a.strip()
                    if not a:
                        continue
                    a2 = a.replace("=", ":").split(":", 2)
                    if len(a2) != 2:
                        valid = False
                        continue
                    k = a2[0].strip().upper()
                    if k not in result:
                        valid = False
                        continue
                    v = a2[1].strip().upper()
                    if v not in ("ACCEPT", "REJECT", "DROP"):
                        valid = False
                        continue
                    result[k] = v

        if not valid:
            raise ValueError("Invalid ReloadPolicy")

        return result

    @staticmethod
    def _unparse_reload_policy(value):
        return ",".join(f"{k}:{v}" for k, v in value.items())



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2010-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import subprocess


def runProg(prog, argv=None, stdin=None):
    if argv is None:
        argv = []

    args = [prog] + argv

    input_string = None
    if stdin:
        with open(stdin, "r") as handle:
            input_string = handle.read().encode()

    env = {"LANG": "C"}
    try:
        process = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            close_fds=True,
            env=env,
        )
    except OSError:
        return (255, "")

    (output, err_output) = process.communicate(input_string)
    output = output.decode("utf-8", "replace")
    return (process.returncode, output)



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2011-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import xml.sax as sax
import os
import io
import shutil

from firewall import config
from firewall.core.io.io_object import (
    IO_Object,
    IO_Object_ContentHandler,
    IO_Object_XMLGenerator,
)
from firewall.core.logger import log
from firewall import errors
from firewall.errors import FirewallError


class IcmpType(IO_Object):
    IMPORT_EXPORT_STRUCTURE = {
        "version": "",  # s
        "short": "",  # s
        "description": "",  # s
        "destination": [""],  # as
    }
    DBUS_SIGNATURE = "(sssas)"
    ADDITIONAL_ALNUM_CHARS = ["_", "-"]
    PARSER_REQUIRED_ELEMENT_ATTRS = {
        "short": None,
        "description": None,
        "icmptype": None,
    }
    PARSER_OPTIONAL_ELEMENT_ATTRS = {
        "icmptype": ["name", "version"],
        "destination": ["ipv4", "ipv6"],
    }

    def __init__(self):
        super(IcmpType, self).__init__()
        self.version = ""
        self.short = ""
        self.description = ""
        self.destination = []

    def cleanup(self):
        self.version = ""
        self.short = ""
        self.description = ""
        del self.destination[:]

    def _check_config(self, config, item, all_config, all_io_objects):
        if item == "destination":
            for destination in config:
                if destination not in ["ipv4", "ipv6"]:
                    raise FirewallError(
                        errors.INVALID_DESTINATION,
                        "'%s' not from {'ipv4'|'ipv6'}" % destination,
                    )


# PARSER


class icmptype_ContentHandler(IO_Object_ContentHandler):
    def startElement(self, name, attrs):
        IO_Object_ContentHandler.startElement(self, name, attrs)
        self.item.parser_check_element_attrs(name, attrs)

        if name == "icmptype":
            if "name" in attrs:
                log.warning("Ignoring deprecated attribute name='%s'" % attrs["name"])
            if "version" in attrs:
                self.item.version = attrs["version"]
        elif name == "short":
            pass
        elif name == "description":
            pass
        elif name == "destination":
            for x in ["ipv4", "ipv6"]:
                if x in attrs and attrs[x].lower() in ["yes", "true"]:
                    self.item.destination.append(str(x))


def icmptype_reader(filename, path):
    icmptype = IcmpType()
    if not filename.endswith(".xml"):
        raise FirewallError(errors.INVALID_NAME, "%s is missing .xml suffix" % filename)
    icmptype.name = filename[:-4]
    icmptype.check_name(icmptype.name)
    icmptype.filename = filename
    icmptype.path = path
    icmptype.builtin = False if path.startswith(config.ETC_FIREWALLD) else True
    icmptype.default = icmptype.builtin
    handler = icmptype_ContentHandler(icmptype)
    parser = sax.make_parser()
    parser.setContentHandler(handler)
    name = "%s/%s" % (path, filename)
    with open(name, "rb") as f:
        source = sax.InputSource(None)
        source.setByteStream(f)
        try:
            parser.parse(source)
        except sax.SAXParseException as msg:
            raise FirewallError(
                errors.INVALID_ICMPTYPE,
                "not a valid icmptype file: %s" % msg.getException(),
            )
    del handler
    del parser
    return icmptype


def icmptype_writer(icmptype, path=None):
    _path = path if path else icmptype.path

    if icmptype.filename:
        name = "%s/%s" % (_path, icmptype.filename)
    else:
        name = "%s/%s.xml" % (_path, icmptype.name)

    if os.path.exists(name):
        try:
            shutil.copy2(name, "%s.old" % name)
        except Exception as msg:
            log.error("Backup of file '%s' failed: %s", name, msg)

    dirpath = os.path.dirname(name)
    if dirpath.startswith(config.ETC_FIREWALLD) and not os.path.exists(dirpath):
        if not os.path.exists(config.ETC_FIREWALLD):
            os.mkdir(config.ETC_FIREWALLD, 0o750)
        os.mkdir(dirpath, 0o750)

    f = io.open(name, mode="wt", encoding="UTF-8")
    handler = IO_Object_XMLGenerator(f)
    handler.startDocument()

    # start icmptype element
    attrs = {}
    if icmptype.version and icmptype.version != "":
        attrs["version"] = icmptype.version
    handler.startElement("icmptype", attrs)
    handler.ignorableWhitespace("\n")

    # short
    if icmptype.short and icmptype.short != "":
        handler.ignorableWhitespace("  ")
        handler.startElement("short", {})
        handler.characters(icmptype.short)
        handler.endElement("short")
        handler.ignorableWhitespace("\n")

    # description
    if icmptype.description and icmptype.description != "":
        handler.ignorableWhitespace("  ")
        handler.startElement("description", {})
        handler.characters(icmptype.description)
        handler.endElement("description")
        handler.ignorableWhitespace("\n")

    # destination
    if icmptype.destination:
        handler.ignorableWhitespace("  ")
        attrs = {}
        for x in icmptype.destination:
            attrs[x] = "yes"
        handler.simpleElement("destination", attrs)
        handler.ignorableWhitespace("\n")

    # end icmptype element
    handler.endElement("icmptype")
    handler.ignorableWhitespace("\n")
    handler.endDocument()
    f.close()
    del handler



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2010-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import os.path
import copy
import random

from firewall.core.prog import runProg
from firewall.core.logger import log
from firewall.functions import (
    tempFile,
    readfile,
    splitArgs,
    check_mac,
    portStr,
    check_single_address,
    check_address,
    normalizeIP6,
)
from firewall import config
from firewall.errors import (
    FirewallError,
    INVALID_PASSTHROUGH,
    INVALID_RULE,
    UNKNOWN_ERROR,
    INVALID_ADDR,
    INVALID_ICMPTYPE,
)
from firewall.core.rich import (
    Rich_Accept,
    Rich_Reject,
    Rich_Drop,
    Rich_Mark,
    Rich_NFLog,
    Rich_Masquerade,
    Rich_ForwardPort,
    Rich_IcmpBlock,
    Rich_Tcp_Mss_Clamp,
)
from firewall.core.base import DEFAULT_ZONE_TARGET
from firewall.core.icmp import ICMP_TYPES, ICMPV6_TYPES
import string

POLICY_CHAIN_PREFIX = ""

BUILT_IN_CHAINS = {
    "security": ["INPUT", "OUTPUT", "FORWARD"],
    "raw": ["PREROUTING", "OUTPUT"],
    "mangle": ["PREROUTING", "POSTROUTING", "INPUT", "OUTPUT", "FORWARD"],
    "nat": ["PREROUTING", "POSTROUTING", "OUTPUT"],
    "filter": ["INPUT", "OUTPUT", "FORWARD"],
}

DEFAULT_REJECT_TYPE = {
    "ipv4": "icmp-host-prohibited",
    "ipv6": "icmp6-adm-prohibited",
}

ICMP = {
    "ipv4": "icmp",
    "ipv6": "ipv6-icmp",
}


def common_reverse_passthrough(args):
    """Reverse valid passthough rule"""

    replace_args = {
        # Append
        "-A": "-D",
        "--append": "--delete",
        # Insert
        "-I": "-D",
        "--insert": "--delete",
        # New chain
        "-N": "-X",
        "--new-chain": "--delete-chain",
    }

    ret_args = args[:]

    for x in replace_args:
        try:
            idx = ret_args.index(x)
        except ValueError:
            continue

        if x in ["-I", "--insert"]:
            # With insert rulenum, then remove it if it is a number
            # Opt at position idx, chain at position idx+1, [rulenum] at
            # position idx+2
            try:
                int(ret_args[idx + 2])
            except ValueError:
                pass
            else:
                ret_args.pop(idx + 2)

        ret_args[idx] = replace_args[x]
        return ret_args

    raise FirewallError(INVALID_PASSTHROUGH, "no '-A', '-I' or '-N' arg")


# ipv ebtables also uses this
#
def common_check_passthrough(args):
    """Check if passthough rule is valid (only add, insert and new chain
    rules are allowed)"""

    args = set(args)
    not_allowed = set(
        [
            "-C",
            "--check",  # check rule
            "-D",
            "--delete",  # delete rule
            "-R",
            "--replace",  # replace rule
            "-L",
            "--list",  # list rule
            "-S",
            "--list-rules",  # print rules
            "-F",
            "--flush",  # flush rules
            "-Z",
            "--zero",  # zero rules
            "-X",
            "--delete-chain",  # delete chain
            "-P",
            "--policy",  # policy
            "-E",
            "--rename-chain",
        ]
    )  # rename chain)
    # intersection of args and not_allowed is not empty, i.e.
    # something from args is not allowed
    if len(args & not_allowed) > 0:
        raise FirewallError(
            INVALID_PASSTHROUGH, "arg '%s' is not allowed" % list(args & not_allowed)[0]
        )

    # args need to contain one of -A, -I, -N
    needed = set(["-A", "--append", "-I", "--insert", "-N", "--new-chain"])
    # empty intersection of args and needed, i.e.
    # none from args contains any needed command
    if len(args & needed) == 0:
        raise FirewallError(INVALID_PASSTHROUGH, "no '-A', '-I' or '-N' arg")


class ip4tables:
    ipv = "ipv4"
    name = "ip4tables"
    policies_supported = True

    def __init__(self, fw):
        self._fw = fw
        self._command = config.COMMANDS[self.ipv]
        self._restore_command = config.COMMANDS["%s-restore" % self.ipv]
        self.wait_option = self._detect_wait_option()
        self.restore_wait_option = self._detect_restore_wait_option()
        self.fill_exists()
        self.available_tables = []
        self.rich_rule_priority_counts = {}
        self.policy_dispatch_index_cache = {}
        self.policy_dispatch_index_cache_ref_count = {}
        self.our_chains = {}  # chains created by firewalld
        self.unique_chain_name_map = (
            {}
        )  # maps long chain names to one with shorter unique ID
        self.random = random.Random()
        self.random.seed(
            version=2, a="string to seed the generator and make test output predictable"
        )

    def fill_exists(self):
        self.command_exists = os.path.exists(self._command)
        self.restore_command_exists = os.path.exists(self._restore_command)

    def __run(self, args):
        # convert to string list
        if self.wait_option and self.wait_option not in args:
            _args = [self.wait_option] + ["%s" % item for item in args]
        else:
            _args = ["%s" % item for item in args]
        log.debug2("%s: %s %s", self.__class__, self._command, " ".join(_args))
        (status, ret) = runProg(self._command, _args)
        if status != 0:
            raise ValueError(
                "'%s %s' failed: %s" % (self._command, " ".join(_args), ret)
            )
        return ret

    def _rule_replace(self, rule, pattern, replacement):
        try:
            i = rule.index(pattern)
        except ValueError:
            return False
        else:
            rule[i : i + 1] = replacement
            return True

    def is_chain_builtin(self, ipv, table, chain):
        return table in BUILT_IN_CHAINS and chain in BUILT_IN_CHAINS[table]

    def build_chain_rules(self, add, table, chain):
        rule = ["-t", table]
        if add:
            rule.append("-N")
        else:
            rule.append("-X")
        rule.append(chain)
        return [rule]

    def build_rule(self, add, table, chain, index, args):
        rule = ["-t", table]
        if add:
            rule += ["-I", chain, str(index)]
        else:
            rule += ["-D", chain]
        rule += args
        return rule

    def check_passthrough(self, args):
        common_check_passthrough(args)

    def reverse_passthrough(self, args):
        return common_reverse_passthrough(args)

    def passthrough_parse_table_chain(self, args):
        table = "filter"
        try:
            i = args.index("-t")
        except ValueError:
            pass
        else:
            if len(args) >= i + 1:
                table = args[i + 1]
        chain = None
        for opt in ["-A", "--append", "-I", "--insert", "-N", "--new-chain"]:
            try:
                i = args.index(opt)
            except ValueError:
                pass
            else:
                if len(args) >= i + 1:
                    chain = args[i + 1]
        return (table, chain)

    def _set_rule_replace_priority(self, rule, priority_counts, token):
        """
        Change something like
          -t filter -I public_IN %%RICH_RULE_PRIORITY%% 123
        or
          -t filter -A public_IN %%RICH_RULE_PRIORITY%% 321
        into
          -t filter -I public_IN 4
        or
          -t filter -I public_IN
        """
        try:
            i = rule.index(token)
        except ValueError:
            pass
        else:
            rule_add = True
            insert = False
            insert_add_index = -1
            rule.pop(i)
            priority = rule.pop(i)
            if not isinstance(priority, int):
                raise FirewallError(
                    INVALID_RULE, "priority must be followed by a number"
                )

            table = "filter"
            for opt in ["-t", "--table"]:
                try:
                    j = rule.index(opt)
                except ValueError:
                    pass
                else:
                    if len(rule) >= j + 1:
                        table = rule[j + 1]
            for opt in ["-A", "--append", "-I", "--insert", "-D", "--delete"]:
                try:
                    insert_add_index = rule.index(opt)
                except ValueError:
                    pass
                else:
                    if len(rule) >= insert_add_index + 1:
                        chain = rule[insert_add_index + 1]

                    if opt in ["-I", "--insert"]:
                        insert = True
                    if opt in ["-D", "--delete"]:
                        rule_add = False

            chain = (table, chain)

            # Add the rule to the priority counts. We don't need to store the
            # rule, just bump the ref count for the priority value.
            if not rule_add:
                if (
                    chain not in priority_counts
                    or priority not in priority_counts[chain]
                    or priority_counts[chain][priority] <= 0
                ):
                    raise FirewallError(
                        UNKNOWN_ERROR, "nonexistent or underflow of priority count"
                    )

                priority_counts[chain][priority] -= 1
            else:
                if chain not in priority_counts:
                    priority_counts[chain] = {}
                if priority not in priority_counts[chain]:
                    priority_counts[chain][priority] = 0

                # calculate index of new rule
                index = 1
                for p in sorted(priority_counts[chain].keys()):
                    if p == priority and insert:
                        break
                    index += priority_counts[chain][p]
                    if p == priority:
                        break

                priority_counts[chain][priority] += 1

                rule[insert_add_index] = "-I"
                rule.insert(insert_add_index + 2, "%d" % index)

    def _set_rule_sort_policy_dispatch(
        self, rule, policy_dispatch_index_cache, policy_dispatch_index_cache_ref_count
    ):
        try:
            i = rule.index("%%POLICY_SORT_KEY%%")
        except ValueError:
            return False

        rule.pop(i)
        sort_tuple = rule.pop(i)

        delete = False
        verb_index = -1

        table = "filter"
        for opt in ["-t", "--table"]:
            try:
                j = rule.index(opt)
            except ValueError:
                continue

            if len(rule) >= j + 1:
                table = rule[j + 1]
        for opt in ["-A", "--append", "-I", "--insert", "-D", "--delete"]:
            try:
                verb_index = rule.index(opt)
            except ValueError:
                continue

            if len(rule) >= verb_index + 1:
                chain = rule[verb_index + 1]

            if opt in ["-D", "--delete"]:
                delete = True

        chain = (table, chain)

        if delete:
            if (
                chain in policy_dispatch_index_cache
                and sort_tuple in policy_dispatch_index_cache[chain]
            ):
                if (
                    chain in policy_dispatch_index_cache_ref_count
                    and sort_tuple in policy_dispatch_index_cache_ref_count[chain]
                ):
                    if policy_dispatch_index_cache_ref_count[chain][sort_tuple] == 1:
                        del policy_dispatch_index_cache_ref_count[chain][sort_tuple]
                        policy_dispatch_index_cache[chain].remove(sort_tuple)
                        return False
                    else:
                        policy_dispatch_index_cache_ref_count[chain][sort_tuple] -= 1
                        return True
                else:  # only ever one, so no ref count created
                    policy_dispatch_index_cache[chain].remove(sort_tuple)
                    return False
            return True
        else:
            if chain not in policy_dispatch_index_cache:
                policy_dispatch_index_cache[chain] = []

            # rule de-duplication
            if sort_tuple in policy_dispatch_index_cache[chain]:
                if chain not in policy_dispatch_index_cache_ref_count:
                    policy_dispatch_index_cache_ref_count[chain] = {}
                if sort_tuple not in policy_dispatch_index_cache_ref_count[chain]:
                    policy_dispatch_index_cache_ref_count[chain][sort_tuple] = 1

                if sort_tuple in policy_dispatch_index_cache_ref_count[chain]:
                    policy_dispatch_index_cache_ref_count[chain][sort_tuple] += 1

                # Tell caller to skip this rule as we already have one
                return True

            policy_dispatch_index_cache[chain].append(sort_tuple)
            policy_dispatch_index_cache[chain].sort()

            index = policy_dispatch_index_cache[chain].index(sort_tuple)

            rule[verb_index] = "-I"
            rule.insert(verb_index + 2, f"{index + 1}")

            return False

    def set_rules(self, rules, log_denied):
        temp_file = tempFile()

        table_rules = {}
        rich_rule_priority_counts = copy.deepcopy(self.rich_rule_priority_counts)
        policy_dispatch_index_cache = copy.deepcopy(self.policy_dispatch_index_cache)
        policy_dispatch_index_cache_ref_count = copy.deepcopy(
            self.policy_dispatch_index_cache_ref_count
        )
        for _rule in rules:
            rule = _rule[:]

            # replace %%REJECT%%
            self._rule_replace(
                rule,
                "%%REJECT%%",
                ["REJECT", "--reject-with", DEFAULT_REJECT_TYPE[self.ipv]],
            )

            # replace %%ICMP%%
            self._rule_replace(rule, "%%ICMP%%", [ICMP[self.ipv]])

            # replace %%LOGTYPE%%
            try:
                i = rule.index("%%LOGTYPE%%")
            except ValueError:
                pass
            else:
                if log_denied == "off":
                    continue
                if log_denied in ["unicast", "broadcast", "multicast"]:
                    rule[i : i + 1] = ["-m", "pkttype", "--pkt-type", log_denied]
                else:
                    rule.pop(i)

            self._set_rule_replace_priority(
                rule, rich_rule_priority_counts, "%%RICH_RULE_PRIORITY%%"
            )
            skip = self._set_rule_sort_policy_dispatch(
                rule, policy_dispatch_index_cache, policy_dispatch_index_cache_ref_count
            )

            if skip:
                continue

            table = "filter"
            # get table form rule
            for opt in ["-t", "--table"]:
                try:
                    i = rule.index(opt)
                except ValueError:
                    pass
                else:
                    if len(rule) >= i + 1:
                        rule.pop(i)
                        table = rule.pop(i)

            # we can not use joinArgs here, because it would use "'" instead
            # of '"' for the start and end of the string, this breaks
            # iptables-restore
            for i, element in enumerate(rule):
                for c in string.whitespace:
                    if c in element and not (
                        element.startswith('"') and element.endswith('"')
                    ):
                        rule[i] = '"%s"' % element

            table_rules.setdefault(table, []).append(rule)

        for table in table_rules:
            rules = table_rules[table]

            temp_file.write("*%s\n" % table)
            for rule in rules:
                temp_file.write(" ".join(rule) + "\n")
            temp_file.write("COMMIT\n")

        temp_file.close()

        stat = os.stat(temp_file.name)
        log.debug2(
            "%s: %s %s",
            self.__class__,
            self._restore_command,
            "%s: %d" % (temp_file.name, stat.st_size),
        )
        args = []
        if self.restore_wait_option:
            args.append(self.restore_wait_option)
        args.append("-n")

        (status, ret) = runProg(self._restore_command, args, stdin=temp_file.name)

        if log.getDebugLogLevel() > 2:
            lines = readfile(temp_file.name)
            if lines is not None:
                i = 1
                for line in lines:
                    log.debug3("%8d: %s" % (i, line), nofmt=1, nl=0)
                    if not line.endswith("\n"):
                        log.debug3("", nofmt=1)
                    i += 1

        os.unlink(temp_file.name)

        if status != 0:
            raise ValueError(
                "'%s %s' failed: %s" % (self._restore_command, " ".join(args), ret)
            )
        self.rich_rule_priority_counts = rich_rule_priority_counts
        self.policy_dispatch_index_cache = policy_dispatch_index_cache
        self.policy_dispatch_index_cache_ref_count = (
            policy_dispatch_index_cache_ref_count
        )

    def set_rule(self, rule, log_denied):
        # replace %%REJECT%%
        self._rule_replace(
            rule,
            "%%REJECT%%",
            ["REJECT", "--reject-with", DEFAULT_REJECT_TYPE[self.ipv]],
        )

        # replace %%ICMP%%
        self._rule_replace(rule, "%%ICMP%%", [ICMP[self.ipv]])

        # replace %%LOGTYPE%%
        try:
            i = rule.index("%%LOGTYPE%%")
        except ValueError:
            pass
        else:
            if log_denied == "off":
                return ""
            if log_denied in ["unicast", "broadcast", "multicast"]:
                rule[i : i + 1] = ["-m", "pkttype", "--pkt-type", log_denied]
            else:
                rule.pop(i)

        rich_rule_priority_counts = copy.deepcopy(self.rich_rule_priority_counts)
        policy_dispatch_index_cache = copy.deepcopy(self.policy_dispatch_index_cache)
        policy_dispatch_index_cache_ref_count = copy.deepcopy(
            self.policy_dispatch_index_cache_ref_count
        )
        self._set_rule_replace_priority(
            rule, rich_rule_priority_counts, "%%RICH_RULE_PRIORITY%%"
        )
        skip = self._set_rule_sort_policy_dispatch(
            rule, policy_dispatch_index_cache, policy_dispatch_index_cache_ref_count
        )

        output = ""
        if not skip:
            output = self.__run(rule)

        self.rich_rule_priority_counts = rich_rule_priority_counts
        self.policy_dispatch_index_cache = policy_dispatch_index_cache
        self.policy_dispatch_index_cache_ref_count = (
            policy_dispatch_index_cache_ref_count
        )
        return output

    def get_available_tables(self, table=None):
        ret = []
        tables = [table] if table else BUILT_IN_CHAINS.keys()
        for table in tables:
            if table in self.available_tables:
                ret.append(table)
            else:
                try:
                    self.__run(["-t", table, "-L", "-n"])
                    self.available_tables.append(table)
                    ret.append(table)
                except ValueError:
                    log.debug1(
                        "%s table '%s' does not exist (or not enough permission to check)."
                        % (self.ipv, table)
                    )

        return ret

    def _detect_wait_option(self):
        wait_option = ""
        ret = runProg(self._command, ["-w", "-L", "-n"])  # since iptables-1.4.20
        log.debug3(
            '%s: %s: probe for wait option (%s): ret=%u, output="%s"',
            self.__class__,
            self._command,
            "-w",
            ret[0],
            ret[1],
        )
        if ret[0] == 0:
            wait_option = "-w"  # wait for xtables lock
            ret = runProg(
                self._command, ["-w10", "-L", "-n"]
            )  # since iptables > 1.4.21
            log.debug3(
                '%s: %s: probe for wait option (%s): ret=%u, output="%s"',
                self.__class__,
                self._command,
                "-w10",
                ret[0],
                ret[1],
            )
            if ret[0] == 0:
                wait_option = "-w10"  # wait max 10 seconds
            log.debug2(
                "%s: %s will be using %s option.",
                self.__class__,
                self._command,
                wait_option,
            )

        return wait_option

    def _detect_restore_wait_option(self):
        temp_file = tempFile()
        temp_file.write("#foo")
        temp_file.close()

        wait_option = ""
        for test_option in ["-w", "--wait=2"]:
            ret = runProg(self._restore_command, [test_option], stdin=temp_file.name)
            log.debug3(
                '%s: %s: probe for wait option (%s): ret=%u, output="%s"',
                self.__class__,
                self._command,
                test_option,
                ret[0],
                ret[1],
            )
            if (
                ret[0] == 0
                and "invalid option" not in ret[1]
                and "unrecognized option" not in ret[1]
            ):
                wait_option = test_option
                break

        log.debug2(
            "%s: %s will be using %s option.",
            self.__class__,
            self._restore_command,
            wait_option,
        )

        os.unlink(temp_file.name)

        return wait_option

    def get_unique_chain_name(self, chain, prefix="", suffix=""):
        iptables_max_len = 28
        if len(chain) <= iptables_max_len:
            return chain
        if chain in self.unique_chain_name_map:
            return self.unique_chain_name_map[chain]

        prefix = (prefix + "_") if prefix else ""
        suffix = ("_" + suffix) if suffix else ""

        while True:
            rand_chain = (
                prefix
                + "".join(
                    self.random.choice(string.ascii_letters + string.digits)
                    for i in range(iptables_max_len - len(prefix) - len(suffix))
                )
                + suffix
            )
            if rand_chain not in self.unique_chain_name_map.values():
                self.unique_chain_name_map[chain] = rand_chain
                return rand_chain

    def build_flush_rules(self):
        self.rich_rule_priority_counts = {}
        self.policy_dispatch_index_cache = {}
        self.policy_dispatch_index_cache_ref_count = {}
        self.our_chains = {}
        self.unique_chain_name_map = {}
        rules = []
        for table in BUILT_IN_CHAINS.keys():
            if not self.get_available_tables(table):
                continue
            # Flush firewall rules: -F
            # Delete firewall chains: -X
            # Set counter to zero: -Z
            for flag in ["-F", "-X", "-Z"]:
                rules.append(["-t", table, flag])
        return rules

    def build_set_policy_rules(self, policy, policy_details):
        rules = []
        _policy = "DROP" if policy == "PANIC" else policy
        for table in BUILT_IN_CHAINS.keys():
            if not self.get_available_tables(table):
                continue
            if table == "nat":
                continue
            for chain in BUILT_IN_CHAINS[table]:
                if table == "filter":
                    p = policy_details[chain]
                    if p == "REJECT":
                        rules.append(["-t", table, "-A", chain, "-j", "REJECT"])
                        p = "DROP"
                else:
                    p = _policy
                rules.append(["-t", table, "-P", chain, p])
        return rules

    def supported_icmp_types(self, ipv=None):
        supported = set()

        if ipv is None or self.ipv == "ipv4":
            supported.update(ICMP_TYPES.keys())
        if ipv is None or self.ipv == "ipv6":
            supported.update(ICMPV6_TYPES.keys())

        return list(supported)

    def build_default_tables(self):
        # nothing to do, they always exist
        return []

    def _build_default_rules_dnat(self, chain):
        dnat_rules = []
        if self._fw._strict_forward_ports:
            self.our_chains["filter"].add(f"{chain}_dnat")
            dnat_rules.append(f"-N {chain}_dnat")
            dnat_rules.append(f"-A {chain}_dnat -j %%REJECT%%")
            dnat_rules.append(f"-A {chain} -m conntrack --ctstate DNAT -j {chain}_dnat")

        return dnat_rules

    def build_default_rules(self, log_denied="off"):
        default_rules = {}

        if self.get_available_tables("security"):
            default_rules["security"] = []
            self.our_chains["security"] = set()
            for chain in BUILT_IN_CHAINS["security"]:
                default_rules["security"].append("-N %s_direct" % chain)
                default_rules["security"].append("-A %s -j %s_direct" % (chain, chain))
                self.our_chains["security"].add("%s_direct" % chain)

        if self.get_available_tables("raw"):
            default_rules["raw"] = []
            self.our_chains["raw"] = set()
            for chain in BUILT_IN_CHAINS["raw"]:
                default_rules["raw"].append("-N %s_direct" % chain)
                default_rules["raw"].append("-A %s -j %s_direct" % (chain, chain))
                self.our_chains["raw"].add("%s_direct" % chain)

                if chain == "PREROUTING":
                    default_rules["raw"].append("-N %s_POLICIES" % (chain))
                    self.our_chains["raw"].update(set(["%s_POLICIES" % (chain)]))
                    default_rules["raw"].append("-A %s -j %s_POLICIES" % (chain, chain))

        if self.get_available_tables("mangle"):
            default_rules["mangle"] = []
            self.our_chains["mangle"] = set()
            for chain in BUILT_IN_CHAINS["mangle"]:
                default_rules["mangle"].append("-N %s_direct" % chain)
                default_rules["mangle"].append("-A %s -j %s_direct" % (chain, chain))
                self.our_chains["mangle"].add("%s_direct" % chain)

                if chain == "PREROUTING":
                    default_rules["mangle"].append("-N %s_POLICIES" % (chain))
                    self.our_chains["mangle"].update(set(["%s_POLICIES" % (chain)]))
                    default_rules["mangle"].append(
                        "-A %s -j %s_POLICIES" % (chain, chain)
                    )

        if self.get_available_tables("nat"):
            default_rules["nat"] = []
            self.our_chains["nat"] = set()
            for chain in BUILT_IN_CHAINS["nat"]:
                default_rules["nat"].append("-N %s_direct" % chain)
                default_rules["nat"].append("-A %s -j %s_direct" % (chain, chain))
                self.our_chains["nat"].add("%s_direct" % chain)

                default_rules["nat"].append("-N %s_POLICIES" % (chain))
                self.our_chains["nat"].update(set(["%s_POLICIES" % (chain)]))
                default_rules["nat"].append("-A %s -j %s_POLICIES" % (chain, chain))

        default_rules["filter"] = []
        self.our_chains["filter"] = set()

        default_rules["filter"].append(
            "-A INPUT -m conntrack --ctstate RELATED,ESTABLISHED{} -j ACCEPT".format(
                "" if self._fw._strict_forward_ports else ",DNAT"
            )
        )
        default_rules["filter"].extend(self._build_default_rules_dnat("INPUT"))

        default_rules["filter"].append("-A INPUT -i lo -j ACCEPT")
        if log_denied != "off":
            default_rules["filter"].append(
                "-A INPUT -m conntrack --ctstate INVALID %%LOGTYPE%% -j LOG --log-prefix 'STATE_INVALID_DROP: '"
            )
        default_rules["filter"].append(
            "-A INPUT -m conntrack --ctstate INVALID -j DROP"
        )
        default_rules["filter"].append("-N INPUT_direct")
        default_rules["filter"].append("-A INPUT -j INPUT_direct")
        self.our_chains["filter"].update(set("INPUT_direct"))
        default_rules["filter"].append("-N INPUT_POLICIES")
        self.our_chains["filter"].update(set("INPUT_POLICIES"))
        default_rules["filter"].append("-A INPUT -j INPUT_POLICIES")
        if log_denied != "off":
            default_rules["filter"].append(
                "-A INPUT %%LOGTYPE%% -j LOG --log-prefix 'FINAL_REJECT: '"
            )
        default_rules["filter"].append("-A INPUT -j %%REJECT%%")

        default_rules["filter"].append(
            "-A FORWARD -m conntrack --ctstate RELATED,ESTABLISHED{} -j ACCEPT".format(
                "" if self._fw._strict_forward_ports else ",DNAT"
            )
        )
        default_rules["filter"].extend(self._build_default_rules_dnat("FORWARD"))

        default_rules["filter"].append("-A FORWARD -i lo -j ACCEPT")
        if log_denied != "off":
            default_rules["filter"].append(
                "-A FORWARD -m conntrack --ctstate INVALID %%LOGTYPE%% -j LOG --log-prefix 'STATE_INVALID_DROP: '"
            )
        default_rules["filter"].append(
            "-A FORWARD -m conntrack --ctstate INVALID -j DROP"
        )
        default_rules["filter"].append("-N FORWARD_direct")
        default_rules["filter"].append("-A FORWARD -j FORWARD_direct")
        self.our_chains["filter"].update(set("FORWARD_direct"))
        default_rules["filter"].append("-N FORWARD_POLICIES")
        self.our_chains["filter"].update(set("FORWARD_POLICIES"))
        default_rules["filter"].append("-A FORWARD -j FORWARD_POLICIES")
        if log_denied != "off":
            default_rules["filter"].append(
                "-A FORWARD %%LOGTYPE%% -j LOG --log-prefix 'FINAL_REJECT: '"
            )
        default_rules["filter"].append("-A FORWARD -j %%REJECT%%")

        default_rules["filter"] += ["-N OUTPUT_direct"]

        default_rules["filter"].append(
            "-A OUTPUT -m conntrack --ctstate RELATED,ESTABLISHED{} -j ACCEPT".format(
                "" if self._fw._strict_forward_ports else ",DNAT"
            )
        )
        default_rules["filter"].extend(self._build_default_rules_dnat("OUTPUT"))

        default_rules["filter"] += [
            "-A OUTPUT -o lo -j ACCEPT",
            "-A OUTPUT -j OUTPUT_direct",
        ]
        self.our_chains["filter"].update(set("OUTPUT_direct"))
        default_rules["filter"].append("-N OUTPUT_POLICIES")
        default_rules["filter"].append("-A OUTPUT -j OUTPUT_POLICIES")
        self.our_chains["filter"].update(set("OUTPUT_POLICIES"))

        final_default_rules = []
        for table in default_rules:
            if table not in self.get_available_tables():
                continue
            for rule in default_rules[table]:
                final_default_rules.append(["-t", table] + splitArgs(rule))

        return final_default_rules

    def get_zone_table_chains(self, table):
        if table == "filter":
            return {"INPUT", "FORWARD"}
        if table == "mangle":
            if "mangle" in self.get_available_tables():
                return {"PREROUTING"}
        if table == "nat":
            if "nat" in self.get_available_tables():
                return {"PREROUTING", "POSTROUTING"}
        if table == "raw":
            if "raw" in self.get_available_tables():
                return {"PREROUTING"}

        return {}

    def _policy_dispatch_sort_key(
        self,
        policy,
        ingress_zone,
        egress_zone,
        ingress_interface,
        ingress_source,
        egress_interface,
        egress_source,
        priority,
        last=False,
        prerouting=False,
        postrouting=False,
        log_denied=False,
    ):
        p_obj = self._fw.policy.get_policy(policy)
        ingress_priority = (
            0
            if ingress_zone == "HOST"
            else self._fw.zone.get_zone(ingress_zone).ingress_priority
        )
        egress_priority = (
            0
            if egress_zone == "HOST"
            else self._fw.zone.get_zone(egress_zone).egress_priority
        )

        ingress_sort_order = 0  # 0 means output chain
        if ingress_source:
            ingress_sort_order = 1
        elif ingress_interface:
            ingress_sort_order = 2
            if postrouting:
                ingress_zone = ""
                ingress_interface = ""

        egress_sort_order = 0  # 0 means input chain
        if egress_source:
            egress_sort_order = 1
        elif egress_interface or (p_obj.derived_from_zone and prerouting):
            egress_sort_order = 2
            if prerouting:
                egress_zone = ""
                egress_interface = ""

        # default zone is always sorted to last as it's a "catch-all"
        if ingress_interface == "+":
            ingress_priority = self._fw.zone.get_zone(ingress_zone).priority_max + 1
        if egress_interface == "+":
            egress_priority = self._fw.zone.get_zone(egress_zone).priority_max + 1

        last_sort_order = 0
        if last:
            if log_denied:
                last_sort_order = 1
            else:
                last_sort_order = 2

        ingress = (
            ingress_priority,
            ingress_sort_order,
            ingress_zone,
            ingress_source,
            ingress_interface,
        )
        egress = (
            egress_priority,
            egress_sort_order,
            egress_zone,
            egress_source,
            egress_interface,
        )
        suffix = (last_sort_order, priority)

        if postrouting:
            return ["%%POLICY_SORT_KEY%%", egress + ingress + suffix + (policy,)]
        else:
            return ["%%POLICY_SORT_KEY%%", ingress + egress + suffix + (policy,)]

    def build_policy_ingress_egress_pair_rules(
        self,
        enable,
        policy,
        table,
        chain,
        ingress_zone,
        egress_zone,
        ingress_interface,
        ingress_source,
        egress_interface,
        egress_source,
        last=False,
    ):
        add_del = {True: "-I", False: "-D"}[enable]
        p_obj = self._fw.policy.get_policy(policy)
        isSNAT = True if (table == "nat" and chain == "POSTROUTING") else False
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX, isSNAT
        )
        _policy_unique = self.get_unique_chain_name(_policy, prefix=POLICY_CHAIN_PREFIX)
        prerouting = True if chain == "PREROUTING" else False
        postrouting = True if chain == "POSTROUTING" else False

        # iptables can not match a destination MAC
        if check_mac(egress_source) or (postrouting and check_mac(ingress_source)):
            return []

        rules = []

        rule = ["-t", table, add_del, f"{chain}_POLICIES"]

        # iptables-legacy cannot match -i in postrouting
        if ingress_interface and ingress_interface != "+" and not postrouting:
            rule.extend(["-i", ingress_interface])
        if egress_interface and egress_interface != "+" and not prerouting:
            rule.extend(["-o", egress_interface])
        if ingress_source:
            rule.extend(self._rule_addr_fragment("-s", ingress_source))
        if egress_source:
            rule.extend(self._rule_addr_fragment("-d", egress_source))

        if not last:
            if _policy == _policy_unique:
                rule.extend(["-j", _policy_unique])
            else:
                rule.extend(
                    [
                        "-j",
                        _policy_unique,
                        "-m",
                        "comment",
                        "--comment",
                        f"human readable jump target: {_policy}",
                    ]
                )
        elif table != "filter" or chain in ["PREROUTING", "OUTPUT"]:
            rule.extend(["-j", "RETURN"])
        elif p_obj.target in [
            DEFAULT_ZONE_TARGET,
            "ACCEPT",
            "REJECT",
            "%%REJECT%%",
            "DROP",
        ]:
            if self._fw.get_log_denied() != "off" and p_obj.target in [
                DEFAULT_ZONE_TARGET,
                "%%REJECT%%",
                "REJECT",
                "DROP",
            ]:
                _rule = rule[:]
                _log_suffix = "DROP" if p_obj.target == "DROP" else "REJECT"

                _rule.extend(
                    [
                        "%%LOGTYPE%%",
                        "-j",
                        "LOG",
                        "--log-prefix",
                        f"{_policy}_{_log_suffix}: ",
                    ]
                )
                _rule.extend(
                    self._policy_dispatch_sort_key(
                        policy,
                        ingress_zone,
                        egress_zone,
                        ingress_interface,
                        ingress_source,
                        egress_interface,
                        egress_source,
                        p_obj.priority,
                        last=True,
                        log_denied=True,
                        postrouting=postrouting,
                        prerouting=prerouting,
                    )
                )
                rules.append(_rule)

            if p_obj.target == DEFAULT_ZONE_TARGET:
                rule.extend(["-j", "REJECT"])
            else:
                rule.extend(["-j", p_obj.target])

        rule.extend(
            self._policy_dispatch_sort_key(
                policy,
                ingress_zone,
                egress_zone,
                ingress_interface,
                ingress_source,
                egress_interface,
                egress_source,
                p_obj.priority,
                last=last,
                postrouting=postrouting,
                prerouting=prerouting,
            )
        )

        rules.append(rule)

        return rules

    def _rule_addr_fragment(self, opt, address, invert=False):
        if address.startswith("ipset:"):
            name = address[6:]
            if opt == "-d":
                opt = "dst"
            else:
                opt = "src"
            flags = ",".join([opt] * self._fw.ipset.get_dimension(name))
            return ["-m", "set", "--match-set", name, flags]
        elif check_mac(address):
            # outgoing can not be set
            if opt == "-d":
                raise FirewallError(INVALID_ADDR, "Can't match a destination MAC.")
            return ["-m", "mac", "--mac-source", address.upper()]
        else:
            if check_single_address("ipv6", address):
                address = normalizeIP6(address)
            elif check_address("ipv6", address):
                addr_split = address.split("/")
                address = normalizeIP6(addr_split[0]) + "/" + addr_split[1]
            return [opt, address]

    def build_policy_chain_rules(self, enable, policy, table, chain):
        add_del_chain = {True: "-N", False: "-X"}[enable]
        add_del_rule = {True: "-A", False: "-D"}[enable]
        isSNAT = True if (table == "nat" and chain == "POSTROUTING") else False
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX, isSNAT=isSNAT
        )
        _policy_unique = self.get_unique_chain_name(_policy, prefix=POLICY_CHAIN_PREFIX)
        p_obj = self._fw.policy.get_policy(policy)

        self.our_chains[table].add(_policy_unique)

        rules = []
        rules.append([add_del_chain, _policy_unique, "-t", table])
        for suffix in ("pre", "log", "deny", "allow", "post"):
            _p = self.get_unique_chain_name(
                f"{_policy}_{suffix}", prefix=POLICY_CHAIN_PREFIX
            )
            self.our_chains[table].add(_p)
            rules.append([add_del_chain, _p, "-t", table])
            if _policy == _policy_unique:
                rules.append([add_del_rule, _policy_unique, "-t", table, "-j", _p])
            else:
                rules.append(
                    [
                        add_del_rule,
                        _policy_unique,
                        "-t",
                        table,
                        "-j",
                        _p,
                        "-m",
                        "comment",
                        "--comment",
                        f"human readable jump target: {_policy}_{suffix}",
                    ]
                )

        if not p_obj.derived_from_zone and table == "filter":
            target = self._fw.policy._policies[policy].target

            if self._fw.get_log_denied() != "off":
                if target in ["REJECT", "%%REJECT%%"]:
                    rules.append(
                        [
                            add_del_rule,
                            _policy_unique,
                            "-t",
                            table,
                            "%%LOGTYPE%%",
                            "-j",
                            "LOG",
                            "--log-prefix",
                            "%s_REJECT: " % _policy,
                        ]
                    )
                elif target == "DROP":
                    rules.append(
                        [
                            add_del_rule,
                            _policy_unique,
                            "-t",
                            table,
                            "%%LOGTYPE%%",
                            "-j",
                            "LOG",
                            "--log-prefix",
                            "%s_DROP: " % _policy,
                        ]
                    )

            if target in ["ACCEPT", "REJECT", "%%REJECT%%", "DROP"]:
                rules.append([add_del_rule, _policy_unique, "-t", table, "-j", target])

        if not enable:
            rules.reverse()

        return rules

    def _rule_limit(self, limit):
        if not limit:
            return []
        s = ["-m", "limit", "--limit", limit.value]
        if limit.burst:
            s += ["--limit-burst", str(limit.burst)]
        return s

    def _rich_rule_chain_suffix(self, rich_rule):
        if type(rich_rule.element) in [
            Rich_Masquerade,
            Rich_ForwardPort,
            Rich_IcmpBlock,
            Rich_Tcp_Mss_Clamp,
        ]:
            # These are special and don't have an explicit action
            pass
        elif rich_rule.action:
            if type(rich_rule.action) not in [
                Rich_Accept,
                Rich_Reject,
                Rich_Drop,
                Rich_Mark,
            ]:
                raise FirewallError(
                    INVALID_RULE, "Unknown action %s" % type(rich_rule.action)
                )
        else:
            raise FirewallError(INVALID_RULE, "No rule action specified.")

        if rich_rule.priority == 0:
            if type(rich_rule.element) in [
                Rich_Masquerade,
                Rich_ForwardPort,
                Rich_Tcp_Mss_Clamp,
            ] or type(rich_rule.action) in [Rich_Accept, Rich_Mark]:
                return "allow"
            elif type(rich_rule.element) in [Rich_IcmpBlock] or type(
                rich_rule.action
            ) in [Rich_Reject, Rich_Drop]:
                return "deny"
        elif rich_rule.priority < 0:
            return "pre"
        else:
            return "post"

    def _rich_rule_chain_suffix_from_log(self, rich_rule):
        if not rich_rule.log and not rich_rule.audit:
            raise FirewallError(INVALID_RULE, "Not log or audit")

        if rich_rule.priority == 0:
            return "log"
        elif rich_rule.priority < 0:
            return "pre"
        else:
            return "post"

    def _rich_rule_priority_fragment(self, rich_rule):
        if rich_rule.priority == 0:
            return []
        return ["%%RICH_RULE_PRIORITY%%", rich_rule.priority]

    def _rich_rule_log(self, policy, rich_rule, enable, table, rule_fragment):
        if not rich_rule.log:
            return []

        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )

        add_del = {True: "-A", False: "-D"}[enable]

        chain_suffix = self._rich_rule_chain_suffix_from_log(rich_rule)
        rule = [
            "-t",
            table,
            add_del,
            self.get_unique_chain_name(
                f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
            ),
        ]
        rule += self._rich_rule_priority_fragment(rich_rule)
        if isinstance(rich_rule.log, Rich_NFLog):
            rule += rule_fragment + ["-j", "NFLOG"]
            if rich_rule.log.group:
                rule += ["--nflog-group", str(rich_rule.log.group)]
            if rich_rule.log.prefix:
                rule += ["--nflog-prefix", "%s" % rich_rule.log.prefix]
            if rich_rule.log.threshold:
                rule += ["--nflog-threshold", str(rich_rule.log.threshold)]
        else:
            rule += rule_fragment + ["-j", "LOG"]
            if rich_rule.log.prefix:
                rule += ["--log-prefix", "%s" % rich_rule.log.prefix]
            if rich_rule.log.level:
                rule += ["--log-level", "%s" % rich_rule.log.level]
        rule += self._rule_limit(rich_rule.log.limit)

        return rule

    def _rich_rule_audit(self, policy, rich_rule, enable, table, rule_fragment):
        if not rich_rule.audit:
            return []

        add_del = {True: "-A", False: "-D"}[enable]

        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )

        chain_suffix = self._rich_rule_chain_suffix_from_log(rich_rule)
        rule = [
            "-t",
            table,
            add_del,
            self.get_unique_chain_name(
                f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
            ),
        ]
        rule += self._rich_rule_priority_fragment(rich_rule)
        rule += rule_fragment
        if isinstance(rich_rule.action, Rich_Accept):
            _type = "accept"
        elif isinstance(rich_rule.action, Rich_Reject):
            _type = "reject"
        elif isinstance(rich_rule.action, Rich_Drop):
            _type = "drop"
        else:
            _type = "unknown"
        rule += ["-j", "AUDIT", "--type", _type]
        rule += self._rule_limit(rich_rule.audit.limit)

        return rule

    def _rich_rule_action(self, policy, rich_rule, enable, table, rule_fragment):
        if not rich_rule.action:
            return []

        add_del = {True: "-A", False: "-D"}[enable]

        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )

        chain_suffix = self._rich_rule_chain_suffix(rich_rule)
        chain = self.get_unique_chain_name(
            f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
        )
        if isinstance(rich_rule.action, Rich_Accept):
            rule_action = ["-j", "ACCEPT"]
        elif isinstance(rich_rule.action, Rich_Reject):
            rule_action = ["-j", "REJECT"]
            if rich_rule.action.type:
                rule_action += ["--reject-with", rich_rule.action.type]
        elif isinstance(rich_rule.action, Rich_Drop):
            rule_action = ["-j", "DROP"]
        elif isinstance(rich_rule.action, Rich_Mark):
            table = "mangle"
            _policy = self._fw.policy.policy_base_chain_name(
                policy, table, POLICY_CHAIN_PREFIX
            )
            chain = self.get_unique_chain_name(
                f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
            )
            rule_action = ["-j", "MARK", "--set-xmark", rich_rule.action.set]
        else:
            raise FirewallError(
                INVALID_RULE, "Unknown action %s" % type(rich_rule.action)
            )

        rule = ["-t", table, add_del, chain]
        rule += self._rich_rule_priority_fragment(rich_rule)
        rule += rule_fragment + rule_action
        rule += self._rule_limit(rich_rule.action.limit)

        return rule

    def _rich_rule_destination_fragment(self, rich_dest):
        if not rich_dest:
            return []

        rule_fragment = []
        if rich_dest.addr:
            if rich_dest.invert:
                rule_fragment.append("!")
            if check_single_address("ipv6", rich_dest.addr):
                rule_fragment += ["-d", normalizeIP6(rich_dest.addr)]
            elif check_address("ipv6", rich_dest.addr):
                addr_split = rich_dest.addr.split("/")
                rule_fragment += [
                    "-d",
                    normalizeIP6(addr_split[0]) + "/" + addr_split[1],
                ]
            else:
                rule_fragment += ["-d", rich_dest.addr]
        elif rich_dest.ipset:
            rule_fragment += ["-m", "set"]
            if rich_dest.invert:
                rule_fragment.append("!")
            flags = self._fw.zone._ipset_match_flags(rich_dest.ipset, "dst")
            rule_fragment += ["--match-set", rich_dest.ipset, flags]

        return rule_fragment

    def _rich_rule_source_fragment(self, rich_source):
        if not rich_source:
            return []

        rule_fragment = []
        if rich_source.addr:
            if rich_source.invert:
                rule_fragment.append("!")
            if check_single_address("ipv6", rich_source.addr):
                rule_fragment += ["-s", normalizeIP6(rich_source.addr)]
            elif check_address("ipv6", rich_source.addr):
                addr_split = rich_source.addr.split("/")
                rule_fragment += [
                    "-s",
                    normalizeIP6(addr_split[0]) + "/" + addr_split[1],
                ]
            else:
                rule_fragment += ["-s", rich_source.addr]
        elif hasattr(rich_source, "mac") and rich_source.mac:
            rule_fragment += ["-m", "mac"]
            if rich_source.invert:
                rule_fragment.append("!")
            rule_fragment += ["--mac-source", rich_source.mac]
        elif hasattr(rich_source, "ipset") and rich_source.ipset:
            rule_fragment += ["-m", "set"]
            if rich_source.invert:
                rule_fragment.append("!")
            flags = self._fw.zone._ipset_match_flags(rich_source.ipset, "src")
            rule_fragment += ["--match-set", rich_source.ipset, flags]

        return rule_fragment

    def build_policy_ports_rules(
        self, enable, policy, proto, port, destination=None, rich_rule=None
    ):
        add_del = {True: "-A", False: "-D"}[enable]
        table = "filter"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )

        rule_fragment = ["-p", proto]
        if port:
            rule_fragment += ["--dport", "%s" % portStr(port)]
        if destination:
            rule_fragment += ["-d", destination]
        if rich_rule:
            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)

        rules = []
        if rich_rule:
            rules.append(
                self._rich_rule_log(policy, rich_rule, enable, table, rule_fragment)
            )
            rules.append(
                self._rich_rule_audit(policy, rich_rule, enable, table, rule_fragment)
            )
            rules.append(
                self._rich_rule_action(policy, rich_rule, enable, table, rule_fragment)
            )
        else:
            rules.append(
                [
                    add_del,
                    self.get_unique_chain_name(
                        f"{_policy}_allow", prefix=POLICY_CHAIN_PREFIX
                    ),
                    "-t",
                    table,
                ]
                + rule_fragment
                + ["-j", "ACCEPT"]
            )

        return rules

    def build_policy_protocol_rules(
        self, enable, policy, protocol, destination=None, rich_rule=None
    ):
        add_del = {True: "-A", False: "-D"}[enable]
        table = "filter"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )

        rule_fragment = ["-p", protocol]
        if destination:
            rule_fragment += ["-d", destination]
        if rich_rule:
            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)

        rules = []
        if rich_rule:
            rules.append(
                self._rich_rule_log(policy, rich_rule, enable, table, rule_fragment)
            )
            rules.append(
                self._rich_rule_audit(policy, rich_rule, enable, table, rule_fragment)
            )
            rules.append(
                self._rich_rule_action(policy, rich_rule, enable, table, rule_fragment)
            )
        else:
            rules.append(
                [
                    add_del,
                    self.get_unique_chain_name(
                        f"{_policy}_allow", prefix=POLICY_CHAIN_PREFIX
                    ),
                    "-t",
                    table,
                ]
                + rule_fragment
                + ["-j", "ACCEPT"]
            )

        return rules

    def build_policy_tcp_mss_clamp_rules(
        self, enable, policy, tcp_mss_clamp_value, destination=None, rich_rule=None
    ):
        chain_suffix = "allow"
        table = "filter"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )
        add_del = {True: "-A", False: "-D"}[enable]

        rule_fragment = ["-p", "tcp", "--tcp-flags", "SYN,RST", "SYN"]
        if rich_rule:
            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
            rule_fragment += self._rich_rule_priority_fragment(rich_rule)
            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)

        if tcp_mss_clamp_value == "pmtu" or not tcp_mss_clamp_value:
            rule_fragment += ["-j", "TCPMSS", "--clamp-mss-to-pmtu"]
        else:
            rule_fragment += ["-j", "TCPMSS", "--set-mss", tcp_mss_clamp_value]

        return [
            [
                "-t",
                "filter",
                add_del,
                self.get_unique_chain_name(
                    f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
                ),
            ]
            + rule_fragment
        ]

    def build_policy_source_ports_rules(
        self, enable, policy, proto, port, destination=None, rich_rule=None
    ):
        add_del = {True: "-A", False: "-D"}[enable]
        table = "filter"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )

        rule_fragment = ["-p", proto]
        if port:
            rule_fragment += ["--sport", "%s" % portStr(port)]
        if destination:
            rule_fragment += ["-d", destination]
        if rich_rule:
            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)

        rules = []
        if rich_rule:
            rules.append(
                self._rich_rule_log(policy, rich_rule, enable, table, rule_fragment)
            )
            rules.append(
                self._rich_rule_audit(policy, rich_rule, enable, table, rule_fragment)
            )
            rules.append(
                self._rich_rule_action(policy, rich_rule, enable, table, rule_fragment)
            )
        else:
            rules.append(
                [
                    add_del,
                    self.get_unique_chain_name(
                        f"{_policy}_allow", prefix=POLICY_CHAIN_PREFIX
                    ),
                    "-t",
                    table,
                ]
                + rule_fragment
                + ["-j", "ACCEPT"]
            )

        return rules

    def build_policy_helper_ports_rules(
        self, enable, policy, proto, port, destination, helper_name, module_short_name
    ):
        table = "raw"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )
        add_del = {True: "-A", False: "-D"}[enable]

        rule = [
            add_del,
            self.get_unique_chain_name(f"{_policy}_allow", prefix=POLICY_CHAIN_PREFIX),
            "-t",
            "raw",
            "-p",
            proto,
        ]
        if port:
            rule += ["--dport", "%s" % portStr(port)]
        if destination:
            rule += ["-d", destination]
        rule += ["-j", "CT", "--helper", module_short_name]

        return [rule]

    def build_zone_forward_rules(
        self, enable, zone, policy, table, interface=None, source=None
    ):
        add_del = {True: "-A", False: "-D"}[enable]
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )

        rules = []
        if interface:
            rules.append(
                [
                    "-t",
                    "filter",
                    add_del,
                    self.get_unique_chain_name(
                        f"{_policy}_allow", prefix=POLICY_CHAIN_PREFIX
                    ),
                    "-o",
                    interface,
                    "-j",
                    "ACCEPT",
                ]
            )
        else:  # source
            # iptables can not match destination MAC
            if check_mac(source):
                return []

            rules.append(
                [
                    "-t",
                    "filter",
                    add_del,
                    self.get_unique_chain_name(
                        f"{_policy}_allow", prefix=POLICY_CHAIN_PREFIX
                    ),
                ]
                + self._rule_addr_fragment("-d", source)
                + ["-j", "ACCEPT"]
            )
        return rules

    def build_policy_masquerade_rules(self, enable, policy, rich_rule=None):
        table = "nat"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX, isSNAT=True
        )

        add_del = {True: "-A", False: "-D"}[enable]

        rule_fragment = []
        if rich_rule:
            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
            rule_fragment += self._rich_rule_priority_fragment(rich_rule)
            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
        else:
            chain_suffix = "allow"

        rules = []
        rules.append(
            [
                "-t",
                "nat",
                add_del,
                self.get_unique_chain_name(
                    f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
                ),
            ]
            + rule_fragment
            + ["!", "-o", "lo", "-j", "MASQUERADE"]
        )

        return rules

    def build_policy_forward_port_rules(
        self, enable, policy, port, protocol, toport, toaddr, rich_rule=None
    ):
        table = "nat"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )
        p_obj = self._fw.policy.get_policy(policy)
        add_del = {True: "-A", False: "-D"}[enable]

        to = ""
        if toaddr:
            if check_single_address("ipv6", toaddr):
                to += "[%s]" % normalizeIP6(toaddr)
            else:
                to += toaddr
        if toport and toport != "":
            to += ":%s" % portStr(toport, "-")

        rules = []
        rule_fragment = []
        port_fragment = ["-p", protocol, "--dport", portStr(port)]
        if rich_rule:
            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
            rule_fragment = self._rich_rule_priority_fragment(rich_rule)
            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)

            rules.append(
                self._rich_rule_log(
                    policy, rich_rule, enable, "nat", rule_fragment + port_fragment
                )
            )
        else:
            chain_suffix = "allow"

        rules.append(
            [
                "-t",
                "nat",
                add_del,
                self.get_unique_chain_name(
                    f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
                ),
            ]
            + rule_fragment
            + port_fragment
            + ["-j", "DNAT", "--to-destination", to]
        )

        if self._fw._strict_forward_ports:
            if "HOST" in p_obj.ingress_zones:
                chain = "OUTPUT"
            elif toaddr:
                chain = "FORWARD"
            else:
                chain = "INPUT"
            rules.append(
                ["-t", "filter", "-I" if enable else "-D", f"{chain}_dnat"]
                + [
                    "-m",
                    "conntrack",
                    "--ctstate",
                    "DNAT",
                    "--ctdir",
                    "ORIGINAL",
                    "--ctorigdstport",
                    portStr(port),
                ]
                + ["-j", "ACCEPT"]
            )

        return rules

    def _icmp_types_fragment(self, icmp_type):
        if self.ipv == "ipv4" and icmp_type in ICMP_TYPES:
            _type, _code, _omit_code = ICMP_TYPES[icmp_type]
            _type_str = str(_type) if _omit_code else str(_type) + "/" + str(_code)
            return ["-m", "icmp", "--icmp-type", _type_str]
        elif self.ipv == "ipv6" and icmp_type in ICMPV6_TYPES:
            _type, _code, _omit_code = ICMPV6_TYPES[icmp_type]
            _type_str = str(_type) if _omit_code else str(_type) + "/" + str(_code)
            return ["-m", "icmp6", "--icmpv6-type", _type_str]
        else:
            raise FirewallError(
                INVALID_ICMPTYPE, f"ICMP type {icmp_type} not supported by {self.name}"
            )

    def build_policy_icmp_block_rules(
        self, enable, policy, ict, rich_rule=None, ipvs=None
    ):
        table = "filter"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )
        add_del = {True: "-A", False: "-D"}[enable]

        if self.ipv == "ipv4":
            proto = ["-p", "icmp"]
            match = self._icmp_types_fragment(ict.name)
        else:
            proto = ["-p", "ipv6-icmp"]
            match = self._icmp_types_fragment(ict.name)

        rules = []
        if self._fw.policy.query_icmp_block_inversion(policy):
            final_chain = self.get_unique_chain_name(
                f"{_policy}_allow", prefix=POLICY_CHAIN_PREFIX
            )
            final_target = "ACCEPT"
        else:
            final_chain = self.get_unique_chain_name(
                f"{_policy}_deny", prefix=POLICY_CHAIN_PREFIX
            )
            final_target = "%%REJECT%%"

        rule_fragment = []
        if rich_rule:
            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
        rule_fragment += proto + match

        if rich_rule:
            rules.append(
                self._rich_rule_log(policy, rich_rule, enable, table, rule_fragment)
            )
            rules.append(
                self._rich_rule_audit(policy, rich_rule, enable, table, rule_fragment)
            )
            if rich_rule.action:
                rules.append(
                    self._rich_rule_action(
                        policy, rich_rule, enable, table, rule_fragment
                    )
                )
            else:
                chain_suffix = self._rich_rule_chain_suffix(rich_rule)
                rules.append(
                    [
                        "-t",
                        table,
                        add_del,
                        self.get_unique_chain_name(
                            f"{_policy}_{chain_suffix}", prefix=POLICY_CHAIN_PREFIX
                        ),
                    ]
                    + self._rich_rule_priority_fragment(rich_rule)
                    + rule_fragment
                    + ["-j", "%%REJECT%%"]
                )
        else:
            if self._fw.get_log_denied() != "off" and final_target != "ACCEPT":
                rules.append(
                    [add_del, final_chain, "-t", table]
                    + rule_fragment
                    + [
                        "%%LOGTYPE%%",
                        "-j",
                        "LOG",
                        "--log-prefix",
                        "%s_ICMP_BLOCK: " % policy,
                    ]
                )
            rules.append(
                [add_del, final_chain, "-t", table]
                + rule_fragment
                + ["-j", final_target]
            )

        return rules

    def build_policy_icmp_block_inversion_rules(self, enable, policy):
        table = "filter"
        _policy = self._fw.policy.policy_base_chain_name(
            policy, table, POLICY_CHAIN_PREFIX
        )
        _policy_unique = self.get_unique_chain_name(_policy, prefix=POLICY_CHAIN_PREFIX)

        rules = []
        rule_idx = 6

        if self._fw.policy.query_icmp_block_inversion(policy):
            ibi_target = "%%REJECT%%"

            if self._fw.get_log_denied() != "off":
                if enable:
                    rule = ["-I", _policy_unique, str(rule_idx)]
                else:
                    rule = ["-D", _policy_unique]

                rule = rule + [
                    "-t",
                    table,
                    "-p",
                    "%%ICMP%%",
                    "%%LOGTYPE%%",
                    "-j",
                    "LOG",
                    "--log-prefix",
                    "%s_ICMP_BLOCK: " % _policy,
                ]
                rules.append(rule)
                rule_idx += 1
        else:
            ibi_target = "ACCEPT"

        if enable:
            rule = ["-I", _policy_unique, str(rule_idx)]
        else:
            rule = ["-D", _policy_unique]
        rule = rule + ["-t", table, "-p", "%%ICMP%%", "-j", ibi_target]
        rules.append(rule)

        return rules

    def build_policy_rich_source_destination_rules(self, enable, policy, rich_rule):
        table = "filter"

        rule_fragment = []
        rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
        rule_fragment += self._rich_rule_source_fragment(rich_rule.source)

        rules = []
        rules.append(
            self._rich_rule_log(policy, rich_rule, enable, table, rule_fragment)
        )
        rules.append(
            self._rich_rule_audit(policy, rich_rule, enable, table, rule_fragment)
        )
        rules.append(
            self._rich_rule_action(policy, rich_rule, enable, table, rule_fragment)
        )

        return rules

    def is_ipv_supported(self, ipv):
        return ipv == self.ipv


class ip6tables(ip4tables):
    ipv = "ipv6"
    name = "ip6tables"

    def build_rpfilter_rules(self, log_denied=False):
        rules = []
        rpfilter_fragment = ["-m", "rpfilter", "--invert", "--validmark"]
        if self._fw._ipv6_rpfilter == "loose":
            rpfilter_fragment += ["--loose"]

        rules.append(
            [
                "-I",
                "PREROUTING",
                "-t",
                "mangle",
            ]
            + rpfilter_fragment
            + [
                "-j",
                "DROP",
            ]
        )
        if log_denied != "off":
            rules.append(
                [
                    "-I",
                    "PREROUTING",
                    "-t",
                    "mangle",
                ]
                + rpfilter_fragment
                + [
                    "-j",
                    "LOG",
                    "--log-prefix",
                    "rpfilter_DROP: ",
                ]
            )
        rules.append(
            [
                "-I",
                "PREROUTING",
                "-t",
                "mangle",
                "-p",
                "ipv6-icmp",
                "--icmpv6-type=neighbour-solicitation",
                "-j",
                "ACCEPT",
            ]
        )  # RHBZ#1575431, kernel bug in 4.16-4.17
        rules.append(
            [
                "-I",
                "PREROUTING",
                "-t",
                "mangle",
                "-p",
                "ipv6-icmp",
                "--icmpv6-type=router-advertisement",
                "-j",
                "ACCEPT",
            ]
        )  # RHBZ#1058505
        return rules

    def build_rfc3964_ipv4_rules(self):
        daddr_list = [
            "::0.0.0.0/96",  # IPv4 compatible
            "::ffff:0.0.0.0/96",  # IPv4 mapped
            "2002:0000::/24",  # 0.0.0.0/8 (the system has no address assigned yet)
            "2002:0a00::/24",  # 10.0.0.0/8 (private)
            "2002:7f00::/24",  # 127.0.0.0/8 (loopback)
            "2002:ac10::/28",  # 172.16.0.0/12 (private)
            "2002:c0a8::/32",  # 192.168.0.0/16 (private)
            "2002:a9fe::/32",  # 169.254.0.0/16 (IANA Assigned DHCP link-local)
            "2002:e000::/19",  # 224.0.0.0/4 (multicast), 240.0.0.0/4 (reserved and broadcast)
        ]

        chain_name = "RFC3964_IPv4"
        self.our_chains["filter"].add(chain_name)

        rules = []
        rules.append(["-t", "filter", "-N", chain_name])
        for daddr in daddr_list:
            rules.append(
                [
                    "-t",
                    "filter",
                    "-I",
                    chain_name,
                    "-d",
                    daddr,
                    "-j",
                    "REJECT",
                    "--reject-with",
                    "addr-unreach",
                ]
            )
            if self._fw._log_denied in ["unicast", "all"]:
                rules.append(
                    [
                        "-t",
                        "filter",
                        "-I",
                        chain_name,
                        "-d",
                        daddr,
                        "-j",
                        "LOG",
                        "--log-prefix",
                        "RFC3964_IPv4_REJECT: ",
                    ]
                )

        # Inject into FORWARD and OUTPUT chains
        index = 4
        if self._fw._strict_forward_ports:
            index += 1
        rules.append(["-t", "filter", "-I", "OUTPUT", str(index), "-j", chain_name])

        index += 1
        if self._fw.get_log_denied() != "off":
            index += 1
        rules.append(
            [
                "-t",
                "filter",
                "-I",
                "FORWARD",
                str(index),
                "-j",
                chain_name,
            ]
        )
        return rules
